(** * Functors between categories and (2,1)-categories *)
Require Import base ord category pcategory.
Set Universe Polymorphism.

(** ** Functors between categories *)

Polymorphic Cumulative Structure functor (A B : cat) := mk_functor
  { omap :> A -> B
  ; map : forall (a b : A), A a b -> B (omap a) (omap b)
  ; _ : forall (a : A), map (cid : A a a) = cid
  ; _ : forall (a b c : A) (f : A a b) (g : A b c),
      map f ∘ map g = map (f ∘ g)
  }.
Arguments mk_functor {A B} omap map _ _.

(** ** Laws *)

Lemma functor_eq (A B : cat) (F G : functor A B) :
  (omap F; map F) = (omap G; map G) :> { omap : A -> B & forall a b, A a b -> B (omap a) (omap b) } ->
  F = G.
Proof.
  destruct F, G => /= p. move: p..1 p..2 => /= {p}.
  destruct 0. cbn. destruct 1. f_equal; exact: pi.
Qed.

Section FunctorLemmas.
  Variables (A B : cat) (F : functor A B).

  Lemma map1 a : map F (cid : A a a) = cid.
  Proof. by case: F a. Qed.

  Lemma mapC a b c (f : A a b) (g : A b c) :
    map F f ∘ map F g = map F (f ∘ g).
  Proof. by case: F a b c f g. Qed.

  Lemma mapC_comp a b c d (f : A a b) (g : A b c) (h : B (F c) d) :
    map F f ∘ (map F g ∘ h) = map F (f ∘ g) ∘ h.
  Proof. by rewrite <-compA, mapC. Qed.
End FunctorLemmas.

Hint Rewrite map1 mapC mapC_comp : cat.

(** ** category of categories *)

Section CAT.
  Implicit Types (A B C D : cat).

  Definition identity_functor A : functor A A :=
    @mk_functor A A idfun (fun _ _ => idfun) (fun _ => erefl) (fun _ _ _ _ _ => erefl).

  Program Definition compose_functor A B C (F : functor A B) (G : functor B C) : functor A C :=
    mk_functor (F \; G) (fun _ _ f => map G (map F f)) _ _.
  Next Obligation. by rewrite !map1. Qed.
  Next Obligation. by rewrite-> !mapC. Qed.

  Lemma functor_idL A B (F : functor A B) :
    compose_functor (identity_functor A) F = F.
  Proof. exact: functor_eq. Qed.

  Lemma functor_idR A B (F : functor A B) :
    compose_functor F (identity_functor B) = F.
  Proof. exact: functor_eq. Qed.

  Lemma functor_compA A B C D (F : functor A B) (G : functor B C) (H : functor C D) :
    compose_functor (compose_functor F G) H = compose_functor F (compose_functor G H).
  Proof. exact: functor_eq. Qed.

  Canonical cat_catMixin := CatMixin functor_idL functor_idR functor_compA.
  Canonical CAT := Eval hnf in Cat cat functor cat_catMixin.
End CAT.

(** ** Functors between (2,1)-categories *)

Polymorphic Cumulative Structure pfunctor (A B : pcat) := mk_pfunctor
  { pfunctor_to_functor :> functor A B
  ; map_proper : forall (a b : A),
      Proper (ord_op ++> ord_op) (@map A B pfunctor_to_functor a b)
  }.
Arguments mk_pfunctor {A B} _ {_}.
Existing Instance map_proper.

Lemma pfunctor_eq (A B : pcat) (F G : pfunctor A B) :
  F = G :> functor A B -> F = G.
Proof.
  destruct F as [F H1], G as [G H2] => /= E. destruct E. f_equal. exact: pi.
Qed.

(** ** Category of (2,1)-categories *)

Section PCAT.
  Implicit Types (A B C D : pcat).

  Global Instance identity_functor_proper A (a b : A) :
    Proper (ord_op ++> ord_op) (@map A A (identity_functor A) a b).
  Proof. move=> /=f g. exact. Qed.

  Definition identity_pfunctor A : pfunctor A A :=
    mk_pfunctor (identity_functor A).

  Global Instance compose_functor_proper A B C (F : pfunctor A B) (G : pfunctor B C) (a b : A) :
    Proper (ord_op ++> ord_op) (@map A C (compose_functor F G) a b).
  Proof. move=> /= f g le_f_g. by rewrite le_f_g. Qed.

  Definition compose_pfunctor A B C (F : pfunctor A B) (G : pfunctor B C) : pfunctor A C :=
    mk_pfunctor (compose_functor F G).

  Lemma pfunctor_idL A B (F : pfunctor A B) :
    compose_pfunctor (identity_pfunctor A) F = F.
  Proof. apply: pfunctor_eq. exact: functor_idL. Qed.

  Lemma pfunctor_idR A B (F : pfunctor A B) :
    compose_pfunctor F (identity_pfunctor B) = F.
  Proof. apply: pfunctor_eq. exact: functor_idR. Qed.

  Lemma pfunctor_compA A B C D (F : pfunctor A B) (G : pfunctor B C) (H : pfunctor C D) :
    compose_pfunctor (compose_pfunctor F G) H = compose_pfunctor F (compose_pfunctor G H).
  Proof. apply: pfunctor_eq. exact: functor_compA. Qed.

  Canonical pcat_catMixin := CatMixin pfunctor_idL pfunctor_idR pfunctor_compA.
  Canonical PCAT := Eval hnf in Cat pcat pfunctor pcat_catMixin.
End PCAT.