(** * Products in Categories *)
Require Import base ord category pcategory functor data.fintype.

(** ** Empty products *)

Polymorphic Cumulative Class has_terminal_object (A : cat) := mk_has_terminal
  { TerminalObj : A
  ; bang : forall a : A, hom a TerminalObj
  ; bang_eta : forall a (f : hom a TerminalObj), f = bang a
  }.

Notation "1" := (@TerminalObj _ _).
Notation "!" := (@bang _ _ _).

(** ** Binary products *)

Polymorphic Cumulative Class has_binary_products (A : cat) := mk_has_binary_products
  { Prod  : A -> A -> A
  ; pi1   : forall (a b : A), hom (Prod a b) a
  ; pi2   : forall (a b : A), hom (Prod a b) b
  ; hpair : forall (a b c : A), hom a b -> hom a c -> hom a (Prod b c)
  ; hpair_pi1 : forall (a b c : A) (f : hom a b) (g : hom a c),
      hpair f g ∘ pi1 b c = f
  ; hpair_pi2 : forall (a b c : A) (f : hom a b) (g : hom a c),
      hpair f g ∘ pi2 b c = g
  ; hpair_eta : forall (a b c : A) (f : hom a (Prod b c)),
      hpair (f ∘ pi1 b c) (f ∘ pi2 b c) = f
  }.

Notation π₁ := (pi1 _ _).
Notation π₂ := (pi2 _ _).
Notation "⟨ f , g ⟩" := (hpair f g)
  (format "⟨ f ,  g ⟩") : cat_scope.

Lemma hpair_pi1_comp (A : cat) `{has_binary_products A} (a b c d : A)
      (f : hom a b) (g : hom a c) (h : hom b d) :
  ⟨f, g⟩ ∘ (π₁ ∘ h) = f ∘ h.
Proof. by rewrite <-compA, hpair_pi1. Qed.

Lemma hpair_pi2_comp (A : cat) `{has_binary_products A} (a b c d : A)
      (f : hom a b) (g : hom a c) (h : hom c d) :
  ⟨f, g⟩ ∘ (π₂ ∘ h) = g ∘ h.
Proof. by rewrite <-compA, hpair_pi2. Qed.

Lemma hpair_eta_id (A : cat) `{has_binary_products A} (a b : A) :
  ⟨π₁,π₂⟩ = cid :> (hom (Prod a b) (Prod a b)).
Proof. rewrite <- hpair_eta. by csimpl. Qed.

Hint Rewrite @hpair_pi1 @hpair_pi2 @hpair_pi1_comp @hpair_pi2_comp
     (*@hpair_eta*) @hpair_eta_id : cat.

Section Products.
  Variable (A : cat).
  Context `{has_binary_products A}.

  Definition hfirst {a b c : A} (f : hom a b) : hom (Prod a c) (Prod b c) :=
    ⟨ π₁ ∘ f, π₂ ⟩.

  Definition hsecond {a b c : A} (f : hom b c) : hom (Prod a b) (Prod a c) :=
    ⟨ π₁, π₂ ∘ f ⟩.

  Lemma hfirst_id a b : @hfirst a a b cid = cid.
  Proof. unfold hfirst. by csimpl. Qed.

  Lemma hsecond_id a b : @hsecond a b b cid = cid.
  Proof. unfold hsecond. by csimpl. Qed.

  Lemma hfirst_comp a b c d (f : hom a b) (g : hom b c) :
    hfirst f ∘ hfirst g = hfirst (f ∘ g) :> hom (Prod _ d) _.
  Proof.
    rewrite <- (hpair_eta (hfirst f ∘ hfirst g)). unfold hfirst. by csimpl.
  Qed.

  Lemma hsecond_comp a b c d (f : hom a b) (g : hom b c) :
    hsecond f ∘ hsecond g = hsecond (f ∘ g) :> hom (Prod d _) _.
  Proof.
    rewrite <- (hpair_eta (hsecond f ∘ hsecond g)). unfold hsecond. by csimpl.
  Qed.

  Lemma hfirst_hsecond a b c d (f : hom a b) (g : hom c d) :
    hfirst f ∘ hsecond g = hsecond g ∘ hfirst f.
  Proof.
    rewrite <- (hpair_eta (hfirst f ∘ hsecond g)), <- hpair_eta.
    unfold hfirst, hsecond. by csimpl.
  Qed.
End Products.

(** ** Finite products *)

Polymorphic Cumulative Class has_products (A : cat) := mk_has_products
  { has_products_to_terminal :> has_terminal_object A
  ; has_products_to_binary :> has_binary_products A
  }.

Section FiniteProducts.
  Variable (A : cat).
  Context `{has_products A}.

  Fixpoint oprod {n : nat} : (fin n -> A) -> A :=
    match n with
    | 0 => fun=> 1
    | m.+1 => fun (f : option (fin m) -> A) => Prod (f None) (oprod (f \o Some))
    end.

  Fixpoint proj {n : nat} : forall {f : fin n -> A} (i : fin n), hom (oprod f) (f i) :=
    match n with
    | 0 => fun _ d => match d with end
    | m.+1 => fun (f : option (fin m) -> A) (d : option (fin m)) =>
      match d with
      | Some i => π₂ ∘ @proj m (Some >> f) i
      | None => π₁
      end
    end.

  Fixpoint hprod {n : nat} : forall {f : fin n -> A} {a : A}, (forall i, hom a (f i)) -> hom a (oprod f) :=
    match n with
    | 0 => fun _ _ _ => !
    | m.+1 => fun (f : option (fin m) -> A) (a : A) (F : forall i, hom a (f i)) =>
                ⟨ F None, @hprod m (Some >> f) a (fun i => F (Some i)) ⟩
    end.

  Lemma proj_hprod {n : nat} {f : fin n -> A} (a : A) (F : forall i : fin n, hom a (f i)) (i : fin n) :
    hprod F ∘ proj i = F i.
  Proof.
    elim: n f F i => //=m ih f F [i|]; csimpl=>//. by apply (ih (Some >> f)).
  Qed.

  Lemma hprod_eta {n : nat} {f : fin n -> A} {a : A} (F : hom a (oprod f)) :
    hprod (fun i => F ∘ proj i) = F.
  Proof.
    elim: n f F => /=[_ F|m ih f F]. by rewrite [F]bang_eta.
    rewrite -{3}[F]hpair_eta. f_equal. rewrite <-(ih (Some >> f) (F ∘ π₂)).
    f_equal. fext=>i. by csimpl.
  Qed.
End FiniteProducts.