Require Import base ord category pcategory functor nat.
Set Universe Polymorphism.

(** ** Presheaf categories *)

Definition PRESHEAF (A : cat) : cat := FUNCTOR (OP A) SET.

Lemma pmap1 (A : cat) (F : PRESHEAF A) (a : A) (x : F a) :
  map F (cid : A a a) x = x.
Proof. by rewrite map1. Qed.

Lemma pmapC (A : cat) (F : PRESHEAF A) (a b c : A) (f : A a b) (g : A b c) (x : F c) :
  map F f (map F g x) = map F (f ∘ g) x.
Proof.
  by have<-: map F g ∘ map F f = map F (f ∘ g) by exact: (@mapC (OP A) SET F _ _ _ g f).
Qed.

Lemma pnatP (A : cat) (F G : PRESHEAF A) (eta : hom F G) (a b : A) (f : A a b) (x : F b) :
  map G f (eta b x) = eta a (map F f x).
Proof.
  change (map G f (eta b x)) with ((eta b ∘ map G f) x).
    by rewrite<-natP.
Qed.

Hint Rewrite pmap1 pmapC pnatP : cat.

(** ** Yoneda lemma *)

Section Yoneda.
  Variable A : cat.

  (* The hom-set functor *)

  Program Definition y (a : A) : PRESHEAF A :=
    mk_functor (fun (b : A^op) => A b a) (fun b c f g => (f : A c b) ∘ g) _ _.
  Next Obligation. fext=> c. exact: idL. Qed.
  Next Obligation. fext=> d /=. by csimpl. Qed.

  Program Definition ymap (a b : A) (f : A a b) : hom (y a) (y b) :=
    mk_nat (fun c g => g ∘ f) _.
  Next Obligation. fext=> g. by csimpl. Qed.

  Lemma ymap1 (a : A) : ymap (@cid A a) = cid.
  Proof. apply: nat_eq; fext=> b /= f. by csimpl. Qed.

  Lemma ymapC (a b c : A) (f : A a b) (g : A b c) :
    ymap f ∘ ymap g = ymap (f ∘ g).
  Proof. apply: nat_eq; fext=> d /= h. by csimpl. Qed.

  Canonical HOM : functor A (PRESHEAF A) := mk_functor y ymap ymap1 ymapC.

  (* The yoneda lemma *)

  Section Lemma.
    Variable (F : PRESHEAF A) (a : A).

    Definition yred (f : hom (y a) F) : F a := f a cid.

    Program Definition yexp (x : F a) : hom (y a) F :=
      mk_nat (fun b (f : A b a) => map F f x) _.
    Next Obligation.
      fext=> /= g. by csimpl.
    Qed.

    Lemma yred_cl (eta : hom (y a) F) :
      yexp (yred eta) = eta.
    Proof.
      apply: nat_eq; fext=> /=b f. rewrite pnatP /=. by csimpl.
    Qed.

    Lemma yred_cr (x : F a) :
      yred (yexp x) = x.
    Proof. cbn. by csimpl. Qed.

    Canonical yred_is_equiv := mk_is_equiv yexp yred_cl yred_cr.
    Canonical YRED : hom (y a) F <~> F a := mk_equiv yred yred_is_equiv.
  End Lemma.

  (* Naturality of YRED *)

  Lemma yred_nat (F : PRESHEAF A) (a b : A) (f : A a b) (eta : hom (y b) F) :
    yred (ymap f ∘ eta) = map F f (yred eta).
  Proof. cbn. unfold yred. csimpl => /=. by csimpl. Qed.

  Lemma yred_natF (F G : PRESHEAF A) (eps : hom F G) (a : A) (eta : hom (y a) F) :
    yred (eta ∘ eps) = eps a (yred eta).
  Proof. by []. Qed.

  (* y is fully faithful *)

  Lemma y_hom_iso (a b : A) :
    hom (y a) (y b) <~> hom a b.
  Proof. exact: YRED. Qed.
End Yoneda.

(** ** (2,1)-presheafs *)

Definition PPRESHEAF (A : pcat) := PFUNCTOR (POP A) POSET.

(* A lot of universe anomalies here: *)

Lemma ppmap1 (A : pcat) (F : PPRESHEAF A) (a : A) (x : F a) :
  @map (POP A) POSET F a a (cid : @hom A a a) x = x.
Proof. by rewrite map1. Qed.

Lemma ppmapC (A : pcat) (F : PPRESHEAF A) (a b c : A) (f : A a b) (g : A b c) (x : F c) :
  @map (POP A) POSET F b a f (@map (POP A) POSET F c b g x) = map F (f ∘ g) x.
Proof.
  by have<-: @map (POP A) POSET F c b g ∘ @map (POP A) POSET F b a f = @map (POP A) POSET F c a (f ∘ g) by
    exact (@mapC (POP A) POSET F c b a g f).
Qed.

Lemma ppnatP (A : pcat) (F G : PPRESHEAF A) (eta : hom F G) (a b : A) (f : A a b) (x : F b) :
  @map (POP A) POSET G b a f (eta b x) = eta a (@map (POP A) POSET F b a f x).
Proof.
  change (@map (POP A) POSET G b a f (eta b x)) with ((eta b ∘ @map (POP A) POSET G b a f) x).
    by rewrite<-natP.
Qed.

(* Anomaly universe undefined *)
(*Hint Rewrite (@ppmap1) (@ppmapC) (@ppnatP) : cat.*)

Global Instance ppmap_proper (A : pcat) (F : PPRESHEAF A) (a b : A) :
  Proper (ord_op ++> eq ++> ord_op) (@map (POP A) POSET F a b).
Proof.
  move=> f g le_f_g x y eqn. subst y. move: x. rewrite -mfun_leE.
  apply map_proper. exact: le_f_g.
Qed.

(** ** (2,1)-yoneda lemma *)

(*Require Import ord.discrete.

Section DiscreteFunctor.
  Program Definition DISCRETE : functor SET POSET :=
    mk_functor (discrete_ordType) (fun A B (f : A -> B) => @mk_mfun (discrete_ordType A) (discrete_ordType B) f _) _ _.
  Next Obligation. by destruct H. Qed.
  Next Obligation. apply: mfun_eq => /=. reflexivity. Qed.
  Next Obligation. apply: mfun_eq => /=. reflexivity. Qed.
End DiscreteFunctor.*)

Section PYoneda.
  Variable A : pcat.

  (* The hom-set pfunctor *)

  Definition mfun_precomp (a b c : A) (f : hom a b) : {mono hom b c -> hom a c}.
    refine (mk_mfun (fun g => f ∘ g) _).
    abstract (by move=> /= g h->).
  Defined.

  Definition mfun_postcomp (a b c : A) (g : hom b c) : {mono hom a b -> hom a c}.
    refine (mk_mfun (fun (f : hom a b) => f ∘ g) _).
    abstract (by move=>/= f1 f2->).
  Defined.
    
  Definition py_functor (a : A) : functor (OP A) POSET.
    refine (mk_functor (fun (b : A^op) => [ordType of @hom A b a]) _ _ _).
    Unshelve. Focus 3. intros b c f. apply: mfun_precomp. exact f.
    - cbn. intros b. apply: mfun_eq. cbn. fext=> f. cbn. exact: idL.
    - cbn. intros b c d f g. apply: mfun_eq. cbn. fext=> h. cbn.
      symmetry. exact: compA.
  Defined.

  Global Instance py_functor_monotone (a : A) (b c : OP A) :
    Proper (ord_op ++> ord_op) (@map _ _ (py_functor a) b c).
  Proof. move=> /= f g le_f_g h /=. apply: comp_proper. exact: le_f_g. reflexivity. Qed.

  Definition py (a : A) : PPRESHEAF A :=
    @mk_pfunctor (POP A) POSET (@py_functor a) (@py_functor_monotone a).

  Definition pymap (a b : A) (f : hom a b) : hom (py a) (py b).
  Proof.
    hnf. refine (mk_nat _ _). Unshelve. Focus 2.
    cbn=> c. hnf. apply mfun_postcomp. exact f.
    abstract (cbn=> c d g; apply: mfun_eq; fext=> h /=; by csimpl).
  Defined.

  Lemma pymap1 (a : A) : pymap (@cid A a) = cid.
  Proof.
    apply: nat_eq; fext=> b /=. apply: mfun_eq. fext=> /= f. exact: idR.
  Qed.

  Lemma pymapC (a b c : A) (f : A a b) (g : A b c) :
    pymap f ∘ pymap g = pymap (f ∘ g).
  Proof.
    apply: nat_eq; fext=> d /=. apply: mfun_eq; fext=> /= h. by csimpl.
  Qed.

  Canonical nat_hom : functor A (PPRESHEAF A) := mk_functor py pymap pymap1 pymapC.

  Global Instance pymap_proper (a b : A) :
    Proper (ord_op ++> ord_op) (@pymap a b).
  Proof.
    move=> /=f g le_f_g x h/=.
    (* universe anomaly rewrite-> le_f_g*)
    apply: comp_proper. reflexivity. exact: le_f_g.
  Qed.

  Canonical PHOM : pfunctor A (PPRESHEAF A) := mk_pfunctor nat_hom.

  (* The yoneda lemma *)
(*
  Section Lemma.
    Variable (F : PPRESHEAF A) (a : A).

    Definition pyred (f : hom (py a) F) : F a := f a cid.

    Definition mfun_map (b : A) (x : F a) : {mono hom b a -> F b}.
      refine (mk_mfun _ _). Unshelve. Focus 2.
      cbn. intros f. apply: (map F). cbn. exact f. exact x.
      move=>/=c d le_c_d. move: x. rewrite -mfun_leE. apply map_proper. exact le_c_d.
    Defined.

    Program Definition pyexp (x : F a) : hom (py a) F :=
      mk_nat (fun b => mfun_map b x) _.
    Next Obligation.
      apply: mfun_eq. fext=> /= g. by rewrite ppmapC.
    Qed.

    Lemma pyred_cl (eta : @nat (POP A) POSET (py a) F) :
      yexp (yred eta) = eta.
    Proof.
      apply: nat_eq; fext=> /=b f. rewrite pnatP /=. by csimpl.
    Qed.

    Lemma yred_cr (x : F a) :
      yred (yexp x) = x.
    Proof. cbn. by csimpl. Qed.

    Canonical yred_is_equiv := mk_is_equiv yexp yred_cl yred_cr.
    Canonical YRED : hom (y a) F <~> F a := mk_equiv yred yred_is_equiv.
  End Lemma.

  (* Naturality of YRED *)

  Lemma yred_nat (F : PRESHEAF A) (a b : A) (f : A a b) (eta : hom (y b) F) :
    yred (ymap f ∘ eta) = map F f (yred eta).
  Proof. cbn. unfold yred. csimpl => /=. by csimpl. Qed.

  Lemma yred_natF (F G : PRESHEAF A) (eps : hom F G) (a : A) (eta : hom (y a) F) :
    yred (eta ∘ eps) = eps a (yred eta).
  Proof. by []. Qed.

  (* y is fully faithful *)

  Lemma y_hom_iso (a b : A) :
    hom (y a) (y b) <~> hom a b.
  Proof. exact: YRED. Qed.*)
End PYoneda.
(** ** Lax presheafs *)
(*
Definition LAX (A : pcat) := LFUNCTOR (POP A) POSET.

Lemma lmap1 (A : pcat) (F : LAX A) (a : A) (x : F a) :
  map F (cid : hom a a) x = x.
Proof. by rewrite map1. Qed.

Lemma lmapC (A : pcat) (F : LAX A) (a b c : A) (f : hom a b) (g : hom b c) (x : F c) :
  map F f (map F g x) = map F (f ∘ g) x.
Proof.
  by have<-: map F g ∘ map F f = map F (f ∘ g) by exact: (@mapC (POP A) POSET F _ _ _ g f).
Qed.

Lemma lnatP (A : pcat) (F G : LAX A) (eta : hom F G) (a b : A) (f : hom a b) (x : F b) :
  eta a (map F f x) ≤ map G f (eta b x).
Proof.
  exact: (@laxP _ _ F G eta b a f).
Qed.

(* Universe errors *)
(*Hint Rewrite lmap1 lmapC lnatP : cat.*)

Section LaxYoneda.
  Variable A : pcat.

  (* The hom-set pfunctor *)

  Definition mfun_precomp (a b c : A) (f : hom a b) : {mono hom b c -> hom a c}.
    refine (mk_mfun (fun g => f ∘ g) _).
    abstract (by move=> /= g h->).
  Defined.

  Definition mfun_postcomp (a b c : A) (g : hom b c) : {mono hom a b -> hom a c}.
    refine (mk_mfun (fun (f : hom a b) => f ∘ g) _).
    abstract (by move=>/= f1 f2->).
  Defined.
    
  Definition py_functor (a : A) : functor (OP A) POSET.
    refine (mk_functor (fun (b : A^op) => [ordType of @hom A b a]) _ _ _).
    Unshelve. Focus 3. intros b c f. apply: mfun_precomp. exact f.
    - cbn. intros b. apply: mfun_eq. cbn. fext=> f. cbn. exact: idL.
    - cbn. intros b c d f g. apply: mfun_eq. cbn. fext=> h. cbn.
      symmetry. exact: compA.
  Defined.

  Global Instance py_functor_monotone (a : A) (b c : OP A) :
    Proper (ord_op ++> ord_op) (@map _ _ (py_functor a) b c).
  Proof. move=> /= f g le_f_g h /=. by rewrite le_f_g. Qed.

  Definition py (a : A) : LAX A :=
    @mk_pfunctor (POP A) POSET (@py_functor a) (@py_functor_monotone a).

  Definition pymap (a b : A) (f : hom a b) : hom (py a) (py b).
  Proof.
    hnf. refine (mk_lax _ _). Unshelve. Focus 2.
    cbn=> c. hnf. apply mfun_postcomp. exact f.
    abstract (cbn=> c d g h; by csimpl).
  Defined.

  Lemma pymap1 (a : A) : pymap (@cid A a) = cid.
  Proof.
    apply: lax_eq; fext=> b /=. apply: mfun_eq. fext=> /= f. exact: idR.
  Qed.

  Lemma pymapC (a b c : A) (f : A a b) (g : A b c) :
    pymap f ∘ pymap g = pymap (f ∘ g).
  Proof.
    apply: lax_eq; fext=> d /=. apply: mfun_eq; fext=> /= h. by csimpl.
  Qed.

  Canonical lax_hom : functor A (LAX A) := mk_functor py pymap pymap1 pymapC.

  Global Instance pymap_proper (a b : A) :
    Proper (ord_op ++> ord_op) (@pymap a b).
  Proof.
    move=> /=f g le_f_g x h/=. by rewrite le_f_g.
  Qed.

  Canonical LHOM : pfunctor A (LAX A) := mk_pfunctor lax_hom.

  (* The yoneda lemma *)

  Section Lemma.
    Variable (F : LAX A) (a : A).

    Definition pyred (f : hom (py a) F) : F a := f a cid.

    Definition mfun_map (b : A) (x : F a) : {mono hom b a -> F b}.
      refine (mk_mfun _ _). Unshelve. Focus 2.
      cbn. intros f. apply: (map F). cbn. exact f. exact x.
      move=>/=c d le_c_d. move: x. rewrite -mfun_leE. apply map_proper. exact le_c_d.
    Defined.

    Definition pyexp (x : F a) : hom (py a) F.
    Proof.
      refine (mk_lax _ _). Unshelve. Focus 2. cbn. intros b. exact (mfun_map b x).
      cbn. intros c d f g. generalize x.
      
      change (map F (f ∘ g) ≤ map F g ∘ map F f).
      intros b. refine (mk_mfun _ _). Unshelve. Focus 2.
      cbn. intros f. apply: (map F). cbn. exact f. exact x.
      cbn=> g h le_g_h /=. move: x. rewrite -mfun_leE.
      apply map_proper. exact le_g_h.
      cbn. intros b c f g. admit.
    Admitted.
      f_equal. generalize (mapC F f g). change (map F f (map F g x)) with ((map F g ∘ map F f) x). move: x. rewrite -mfun_leE.
      (map_proper F le_g_h). rewrite le_g_h.
      exact (map F f x).
    
    :=
      mk_nat (fun b (f : A b a) => map F f x) _.
    Next Obligation.
      fext=> /= g. by csimpl.
    Qed.

    Lemma yred_cl (eta : hom (y a) F) :
      yexp (yred eta) = eta.
    Proof.
      apply: nat_eq; fext=> /=b f. rewrite pnatP /=. by csimpl.
    Qed.

    Lemma yred_cr (x : F a) :
      yred (yexp x) = x.
    Proof. cbn. by csimpl. Qed.

    Canonical yred_is_equiv := mk_is_equiv yexp yred_cl yred_cr.
    Canonical YRED : hom (y a) F <~> F a := mk_equiv yred yred_is_equiv.
  End Lemma.

  (* Naturality of YRED *)

  Lemma yred_nat (F : PRESHEAF A) (a b : A) (f : A a b) (eta : hom (y b) F) :
    yred (ymap f ∘ eta) = map F f (yred eta).
  Proof. cbn. unfold yred. csimpl => /=. by csimpl. Qed.

  Lemma yred_natF (F G : PRESHEAF A) (eps : hom F G) (a : A) (eta : hom (y a) F) :
    yred (eta ∘ eps) = eps a (yred eta).
  Proof. by []. Qed.

  (* y is fully faithful *)

  Lemma y_hom_iso (a b : A) :
    hom (y a) (y b) <~> hom a b.
  Proof. exact: YRED. Qed.
End Yoneda.
*)