Require Import base ord.

Module Quantale.
  Record mixin_of (T : clat) := Mixin {
    one : T;
    star : T -> T -> T;
    wand : T -> T -> T;
    _ : forall x y z : T, (star x y ≤ z) <-> (x ≤ wand y z);
    _ : forall x y : T, star x y ≤ star y x;
    _ : forall x y z : T, star (star x y) z ≤ star x (star y z);
    _ : forall x : T, star one x = x
  }.

  Section ClassDef.
    Record class_of (T : Type) := Class {
      base : CLat.class_of T;
      mixin : mixin_of (CLat.Pack base T)
    }.
    Local Coercion base : class_of >-> CLat.class_of.

    Structure type := Pack {sort; _ : class_of sort; _ : Type }.
    Local Coercion sort : type >-> Sortclass.

    Variable (T : Type) (cT : type).
    Definition class := let: Pack _ c _ as cT' := cT return class_of cT' in c.
    Definition clone c of phant_id class c := @Pack T c T.
    Let xT := let: Pack T _ _ := cT in T.
    Notation xclass := (class : class_of xT).

    Definition pack b0 (m0 : mixin_of (@CLat.Pack T b0 T)) :=
      [find pt | CLat.sort pt ~ T | "is not a clat"]
      [find pe | CLat.class pt ~ pe]
      [find m  | m ~ m0 | "is not the right mixin"]
      @Pack T (@Class T pe m) T.

    Definition proType := @Pro.Pack cT xclass xT.
    Definition ordType := @Ord.Pack cT xclass xT.
    Definition clat := @CLat.Pack cT xclass xT.
  End ClassDef.

  Module Exports.
    Coercion base : class_of >-> CLat.class_of.
    Coercion mixin : class_of >-> mixin_of.
    Coercion sort : type >-> Sortclass.

    Coercion proType : type >-> Pro.type.
    Canonical proType.

    Coercion ordType : type >-> Ord.type.
    Canonical ordType.

    Coercion clat : type >-> CLat.type.
    Canonical clat.

    Notation quantale := type.
    Notation QuantaleMixin := Mixin.
    Notation Quantale T m := (@pack T _ m _ id _ id _ id).

    Notation "[ 'quantaleMixin' 'of' T ]" := (class _ : mixin_of T)
      (at level 0, format "[ 'quantaleMixin'  'of'  T ]") : form_scope.
    Notation "[ 'quantale' 'of' T 'for' C ]" := (@clone T C _ idfun)
      (at level 0, format "[ 'quantale'  'of'  T  'for'  C ]") : form_scope.
    Notation "[ 'quantale' 'of' T ]" := (@clone T _ _ id)
      (at level 0, format "[ 'quantale'  'of'  T ]") : form_scope.
  End Exports.
End Quantale.
Export Quantale.Exports.

(** ** Operations *)

Definition I {T : quantale} : T :=
  Quantale.one (Quantale.class T).

Definition wand {T : quantale} (x y : T) : T :=
  Quantale.wand (Quantale.class T) x y.

Definition star {T : quantale} (x y : T) : T :=
  Quantale.star (Quantale.class T) x y.

Notation "x ⊗ y" := (star x y)
  (at level 80, right associativity) : ord_scope.
Notation "x -∘ y" := (wand x y)
  (at level 98, right associativity) : ord_scope.

(* Classical linear logic *)
Module CLL.
  Record mixin_of (T : quantale) := Mixin {
    zero : T;
    _ : forall x, (x -∘ zero) -∘ zero ≤ x
  }.

  Section ClassDef.
    Record class_of (T : Type) := Class {
      base : Quantale.class_of T;
      mixin : mixin_of (Quantale.Pack base T)
    }.
    Local Coercion base : class_of >-> Quantale.class_of.

    Structure type := Pack {sort; _ : class_of sort; _ : Type }.
    Local Coercion sort : type >-> Sortclass.

    Variable (T : Type) (cT : type).
    Definition class := let: Pack _ c _ as cT' := cT return class_of cT' in c.
    Definition clone c of phant_id class c := @Pack T c T.
    Let xT := let: Pack T _ _ := cT in T.
    Notation xclass := (class : class_of xT).

    Definition pack b0 (m0 : mixin_of (@Quantale.Pack T b0 T)) :=
      [find pt | Quantale.sort pt ~ T | "is not a quantale"]
      [find pe | Quantale.class pt ~ pe]
      [find m  | m ~ m0 | "is not the right mixin"]
      @Pack T (@Class T pe m) T.

    Definition proType := @Pro.Pack cT xclass xT.
    Definition ordType := @Ord.Pack cT xclass xT.
    Definition clat := @CLat.Pack cT xclass xT.
    Definition quantale := @Quantale.Pack cT xclass xT.
  End ClassDef.

  Module Exports.
    Coercion base : class_of >-> Quantale.class_of.
    Coercion mixin : class_of >-> mixin_of.
    Coercion sort : type >-> Sortclass.

    Coercion proType : type >-> Pro.type.
    Canonical proType.

    Coercion ordType : type >-> Ord.type.
    Canonical ordType.

    Coercion clat : type >-> CLat.type.
    Canonical clat.

    Coercion quantale : type >-> Quantale.type.
    Canonical quantale.

    Notation cll := type.
    Notation CLLMixin := Mixin.
    Notation CLL T m := (@pack T _ m _ id _ id _ id).

    Notation "[ 'cllMixin' 'of' T ]" := (class _ : mixin_of T)
      (at level 0, format "[ 'cllMixin'  'of'  T ]") : form_scope.
    Notation "[ 'cll' 'of' T 'for' C ]" := (@clone T C _ idfun)
      (at level 0, format "[ 'cll'  'of'  T  'for'  C ]") : form_scope.
    Notation "[ 'cll' 'of' T ]" := (@clone T _ _ id)
      (at level 0, format "[ 'cll'  'of'  T ]") : form_scope.
  End Exports.
End CLL.
Export CLL.Exports.

Section QuantaleLaws.
  Variable (T : quantale).
  Implicit Types (x y z : T).

  (** ** Commutative monoid structure *)

  Lemma starC x y : (x ⊗ y) = (y ⊗ x).
  Proof.
    wlog suff h: x y / x ⊗ y ≤ y ⊗ x. exact: le_eq.
      by case: T x y => ?[?[]].
  Qed.

  Lemma starA x y z : ((x ⊗ y) ⊗ z) = (x ⊗ (y ⊗ z)).
  Proof.
    wlog suff h: x y z / (x ⊗ y) ⊗ z ≤ x ⊗ (y ⊗ z).
    apply: le_eq => //. by rewrite [_ ⊗ z]starC starC h starC [x ⊗ _]starC.
      by case: T x y z => ?[?[]].
  Qed.

  Lemma star1x x : (I ⊗ x) = x.
  Proof. by case: T x => ?[?[]]. Qed.

  Lemma starx1 x : (x ⊗ I) = x.
  Proof. by rewrite starC star1x. Qed.

  (** ** Adjunction (monoidal closed structure) *)

  Lemma star_adj x y z : (x ⊗ y ≤ z) <-> (x ≤ y -∘ z).
  Proof. by case: T x y z => ?[?[]]. Qed.

  Global Instance star_is_adjunction x : is_adjunction (star x) (wand x).
  Proof.
    move=> y z. by rewrite starC star_adj.
  Qed.

  Global Instance star_proper :
    Proper (ord_op ++> ord_op ++> ord_op) (@star T).
  Proof.
    move=> x1 x2 l1 y1 y2 l2. rewrite star_adj l1 -star_adj.
    by rewrite starC star_adj l2 -star_adj starC.
  Qed.

  Global Instance wand_proper :
    Proper (ord_op --> ord_op ++> ord_op) (@wand T).
  Proof.
    move=> x1 x2 l1 y1 y2 l2. rewrite -star_adj -l2 starC star_adj.
      by rewrite l1 -star_adj starC star_adj.
  Qed.

  (* Commuting star and lubs *)

  Lemma starxB x : (x ⊗ ⊥) = ⊥.
  Proof. exact: adjB. Qed.

  Lemma starBx x : (⊥ ⊗ x) = ⊥.
  Proof. by rewrite starC starxB. Qed.

  Lemma starxJ x y z : (x ⊗ (y ∨ z)) = ((x ⊗ y) ∨ (x ⊗ z)).
  Proof. exact: adjJ. Qed.

  Lemma starJx x y z : ((x ∨ y) ⊗ z) = ((x ⊗ z) ∨ (y ⊗ z)).
  Proof. by rewrite starC starxJ starC [_ ⊗ y]starC. Qed.

  Lemma starxE I (F : I -> T) x : (x ⊗ sup F) = (∃ i, x ⊗ F i).
  Proof. exact: adjE. Qed.

  Lemma starEx I (F : I -> T) x : (sup F ⊗ x) = (∃ i, F i ⊗ x).
  Proof. rewrite starC starxE. f_equal; fext=> i. by rewrite starC. Qed.

  Lemma starxEc I (P : I -> Prop) (F : I -> T) x :
    (x ⊗ ∃ i | P i, F i) = (∃ i | P i, x ⊗ F i).
  Proof. rewrite adjE. f_equal; fext=> i. by rewrite adjE. Qed.

  Global Instance supguard_proper (T : clat) :
    Proper (impl ++> ord_op ++> ord_op) (@supguard T).
  Proof.
    move=> P Q h x y le_x_y. focus=> /h q. exact: exI q _.
  Qed.

  (*Global Instance subrelation_pext :
    subrelation iff eq.
  Proof. move=> P Q []. exact: pext. Qed.*)

  Global Instance supguard_proper_eq (T : clat) :
    Proper (iff ++> eq ++> eq) (@supguard T).
  Proof.
    move=> P Q h x y ->. f_equal. apply: pext; apply h.
  Qed.

  (*Global Instance ord_op_eq_subrelation (T : proType) :
    subrelation (@eq T) (@ord_op T).
  Proof. by move=> x y ->. Qed.

  Global Instance impl_eq_subrelation :
    subrelation (@eq Prop) impl.
  Proof. by move=> P Q ->. Qed.*)

  Global Instance sup_proper_le (T : clat) (I : Type) :
    Proper (pointwise_relation I ord_op ==> ord_op) (@sup T I).
  Proof. move=> F G e. focus=> i. apply: exI (i) _. exact: e. Qed.

  Global Instance sup_proper_eq (T : clat) (I : Type) :
    Proper (pointwise_relation I eq ==> eq) (@sup T I).
  Proof.
    move=> F G e. apply: le_eq; apply: sup_proper_le => i; by rewrite (e i).
  Qed.

  Lemma starExc I (P : I -> Prop) (F : I -> T) x :
    ((∃ i | P i, F i) ⊗ x) = ∃ i | P i, F i ⊗ x.
  Proof.
    rewrite starC starxEc. by setoid_rewrite (starC x).
  Qed.

  (* Commuting wand and glbs *)

  Lemma wandxT x : (x -∘ ⊤) = ⊤.
  Proof. exact: adjT. Qed.

  Lemma wandxM x y z : (x -∘ (y ∧ z)) = ((x -∘ y) ∧ (x -∘ z)).
  Proof. exact: adjM. Qed.

  Lemma wandxA I (F : I -> T) x : (x -∘ inf F) = (∀ i, x -∘ F i).
  Proof. exact: adjA. Qed.

  Lemma wandxAc I (P : I -> Prop) (F : I -> T) x :
    (x -∘ ∀ i | P i, F i) = (∀ i | P i, x -∘ F i).
  Proof. rewrite wandxA. f_equal; fext=> i. exact: wandxA. Qed.

  (* Combining wand with star *)

  Lemma starxW x y : x ⊗ (x -∘ y) ≤ y.
  Proof. exact: adj_counit. Qed.

  Lemma starWx x y : (x -∘ y) ⊗ x ≤ y.
  Proof. by rewrite starC starxW. Qed.

  Lemma wandxW x y z : (x -∘ (y -∘ z)) = ((x ⊗ y) -∘ z).
  Proof.
    apply: le_eq. by rewrite -star_adj -starA !starWx.
    rewrite -!star_adj starA. exact: starWx.
  Qed.

  Lemma wand1x x : (I -∘ x) = x.
  Proof.
    apply: le_eq. rewrite -[_ -∘ _]starx1. exact: starWx.
      by rewrite -star_adj starx1.
  Qed.

  Lemma sep_cut a b x y : a ≤ x -∘ y -> b ≤ x -> a ⊗ b ≤ y.
  Proof. move=> -> ->. exact: starWx. Qed.

  (* Star and infima *)

  Lemma starxT x : x ≤ x ⊗ ⊤.
  Proof. rewrite -{1}[x]starx1. apply: mono. exact: topI. Qed.

  Lemma starTx x : x ≤ ⊤ ⊗ x.
  Proof. by rewrite starC -starxT. Qed.

  Lemma starxM x y z : x ⊗ (y ∧ z) ≤ (x ⊗ y) ∧ (x ⊗ z).
  Proof. exact: monoM. Qed.

  Lemma starMx x y z : (x ∧ y) ⊗ z ≤ (x ⊗ z) ∧ (y ⊗ z).
  Proof. by rewrite starC starxM starC [_ ⊗ y]starC. Qed.

  Lemma starxI I (F : I -> T) x : x ⊗ inf F ≤ ∀ i, x ⊗ F i.
  Proof. exact: monoA. Qed.

  Lemma starIx I (F : I -> T) x : inf F ⊗ x ≤ ∀ i, F i ⊗ x.
  Proof.
    rewrite starC starxI. focus=> i. apply: allE (i) _. by rewrite starC.
  Qed.

  (*Lemma starxIc I (P : I -> Prop) (F : I -> T) x :
    x ** \inf_(i | P i) F i <= \inf_(i | P i) (x ** F i).
  Proof. exact: mono_infc. Qed.

  Lemma starIxc I (P : I -> Prop) (F : I -> T) x :
    (\inf_(i | P i) F i) ** x <= \inf_(i | P i) (F i ** x).
  Proof. rewrite starC starxIc. by setoid_rewrite (starC x). Qed.*)

  (* Wand and suprema *)

  Lemma wandBx x : (⊥ -∘ x) = ⊤.
  Proof.
    apply: top_eq. apply/star_adj. rewrite starxB. exact: botE.
  Qed.

  (* TODO *)

  (* Affine modalities *)

  Definition is_affine x := x ≤ I.
  Existing Class is_affine.

  Definition aff x := I ∧ x.

  Lemma affE x {H : is_affine x} :
    aff x = x.
  Proof.
    apply: le_eq. exact: meetEr. exact: meetI.
  Qed.

  Global Instance is_affine_bot : is_affine bot.
  Proof. exact: botE. Qed.

  Global Instance is_affine_I : is_affine I.
  Proof. exact: le_refl. Qed.

  Global Instance is_affine_aff x : is_affine (aff x).
  Proof. exact: meetEl. Qed.

  Global Instance is_affine_meetL x y (h : is_affine x) :
    is_affine (x ∧ y).
  Proof. exact: meetEl. Qed.

  Global Instance is_affine_meetR x y (h : is_affine y) :
    is_affine (x ∧ y).
  Proof. exact: meetEr. Qed.

  Global Instance is_affine_join x y (h1 : is_affine x) (h2 : is_affine y) :
    is_affine (x ∨ y).
  Proof. exact: joinE. Qed.

  Global Instance is_affine_sup I (F : I -> T) :
    (forall i, is_affine (F i)) -> is_affine (sup F).
  Proof.
    move=> h. rewrite/is_affine. focus. exact: h.
  Qed.

  (* FIXME: should this really be an instance? *)
  Global Instance is_affine_inf I (F : I -> T) (i : I) :
    is_affine (F i) -> is_affine (inf F).
  Proof.
    move=> h. rewrite/is_affine. exact: allE (i) _.
  Qed.

  Global Instance is_affine_star x y (h1 : is_affine x) (h2 : is_affine y) :
    is_affine (x ⊗ y).
  Proof. by rewrite/is_affine h1 star1x. Qed.

  Lemma star_weaken x y {h : is_affine x} :
    x ⊗ y ≤ y.
  Proof. by rewrite h star1x. Qed.

  (* Affine idempotents *)

  Definition is_persistent (x : T) :=
    is_affine x /\ (x ≤ x ⊗ x).
  Existing Class is_persistent.

  Definition box (x : T) : T :=
    ∃ y | is_persistent y /\ (y ≤ x), y.

  Global Instance box_is_persistent (x : T) :
    is_persistent (box x).
  Proof.
    split.
    - rewrite/is_affine/box. by focus=> i [[]]//.
    - rewrite{1}/box. focus=> i [[a p] q]. rewrite p.
      suff: i ≤ box x by move->. exact: exIc (i) _ _.
  Qed.

  Global Instance is_persistent_is_affine (x : T) :
    is_persistent x -> is_affine x.
  Proof. firstorder. Qed.

  Lemma dup (x : T) {h : is_persistent x} :
    x = (x ⊗ x).
  Proof.
    case: h => [p q]. apply: le_eq => //. exact: star_weaken.
  Qed.

  Lemma boxI (x y : T) :
    x ≤ I -> x ≤ x ⊗ x -> x ≤ y ->
    x ≤ box y.
  Proof.
    move=> h1 h2 h3. apply: exIc (x) _ le_refl. by split; [split|].
  Qed.

  Lemma star_to_meet (x y : T) {h1 : is_affine x} {h2 : is_affine y} :
    x ⊗ y ≤ x ∧ y.
  Proof.
    focus; [rewrite starC|]; exact: star_weaken.
  Qed.

  Global Instance bot_is_persistent :
    is_persistent ⊥.
  Proof. split. exact _. exact: botE. Qed.

  Global Instance one_is_persistent :
    is_persistent I.
  Proof. split. exact _. by rewrite star1x. Qed.

  Global Instance join_is_persistent (x y : T) {hx : is_persistent x} {hy : is_persistent y} :
    is_persistent (x ∨ y).
  Proof.
    split. exact _. rewrite starJx !starxJ. rewrite -!dup.
    focus. by do 2 apply: joinIl. by do 2 apply: joinIr.
  Qed.

  Global Instance sup_is_persistent I (F : I -> T) {h : forall i, is_persistent (F i)} :
    is_persistent (sup F).
  Proof.
    split. exact _. focus=> i. rewrite [F i]dup. apply: star_proper; exact: exI.
  Qed.

  Global Instance star_is_persistent (x y : T) {hx : is_persistent x} {hy : is_persistent y} :
    is_persistent (x ⊗ y).
  Proof.
    split. exact _. by rewrite starA [y ⊗ _]starC !starA -dup -starA -dup.
  Qed.

  Lemma star_persistent_meet (x y z : T)
    {hx : is_persistent x} {hy : is_persistent y} {hz : is_persistent z} :
    ((x ≤ y) /\ (x ≤ z)) <-> (x ≤ y ⊗ z).
  Proof.
    split.
    - move=> [le_x_y le_x_z]. by rewrite [x]dup {1}le_x_y le_x_z.
    - move->. rewrite star_to_meet. split. exact: meetEl. exact: meetEr.
  Qed.

  Global Instance box_is_kernel :
    is_kernel box.
  Proof.
    move=> x y. split=> le.
    - rewrite le. rewrite/box. by focus=> z [pz le_z_x].
    - rewrite/box. focus=> z [pz le_z_by]. apply: exIc (box x) _ _.
      split. exact _. exact: le. exact: exIc (z) _ _.
  Qed.

  Lemma persistentP (x : T) {h : is_persistent x} :
    box x = x.
  Proof.
    apply: kernel_stable. exact: exIc (x) _ _.
  Qed.

  Lemma box_star (x y : T) :
    box (x ∧ y) = (box x ⊗ box y).
  Proof.
    apply: le_eq.
    - apply (@star_persistent_meet (box (x ∧ y)) (box x) (box y)). exact _. exact _. exact _.
      split. apply: mono. exact: meetEl. apply: mono. exact: meetEr.
    - rewrite -[_ ⊗ _]persistentP. apply/kernelP. focus.
      + rewrite starC star_weaken kernel_idem. exact: kernel_dec.
      + rewrite star_weaken kernel_idem. exact: kernel_dec.
  Qed.

  Lemma box_top :
    box ⊤ = I.
  Proof.
    rewrite -[I]persistentP. apply: kernel_eq; last by exact: topI.
    rewrite/box. by focus=> i [[]].
  Qed.

  Lemma box_sup J (F : J -> T) :
    box (∃ i, box (F i)) = ∃ i, box (F i).
  Proof.
    apply: kernel_lub. instantiate (1 := F). exact: ex_axiom.
  Qed.

  Lemma ex_elim_int J (F : J -> T) (x : T) :
    box ((∃ i, box (F i)) ∧ x) = ∃ i, box (F i ∧ x).
  Proof.
    rewrite box_star. transitivity (∃ i, box (F i) ⊗ box x); last first.
    f_equal; fext=> i. by rewrite box_star.
    rewrite -starEx. f_equal. by rewrite box_sup.
  Qed.

  (* Dualizing *)

  (* Definition dual (x : T) : T := (x -∘ ⊥) -∘ ⊥. *)

  (* Lemma dual_closure : is_closure dual. *)
  (* Proof. *)
  (*   move=> x y. split=> [<-|le_x_dy]. apply/star_adj. by rewrite starxW. *)
  (*   apply/star_adj. rewrite/dual. rewrite le_x_dy. apply: sep_cut le_refl. *)
  (*   apply: wand_proper => //. rewrite/flip. apply/star_adj. *)
  (*   rewrite/dual. by rewrite starxW. *)
  (* Qed. *)

  (* Lemma dual_neg x : (dual x -∘ ⊥) = (x -∘ ⊥). *)
  (* Proof. *)
  (*   apply: le_eq. apply: wand_proper => //=. rewrite/flip. exact/dual_closure. *)
  (*   apply/star_adj. by rewrite starxW. *)
  (* Qed. *)

  (* Lemma dual_star_strong x y : x ⊗ dual y ≤ dual (x ⊗ y). *)
  (* Proof. *)
  (*   rewrite{2}/dual -star_adj. rewrite starA [dual y ⊗ _]starC -starA. *)
  (*   rewrite star_adj. rewrite dual_neg. rewrite starC star_adj. *)
  (*   by rewrite wandxW. *)
  (* Qed. *)

  (* Lemma dual_star x y : dual x ⊗ dual y ≤ dual (x ⊗ y). *)
  (* Proof. *)
  (*   rewrite dual_star_strong starC. *)
  (*   rewrite{1}/dual dual_star_strong. by rewrite dual_neg starC. *)
  (* Qed. *)

(*
  Definition affine (p : T) : T := I \cap p.
  Definition effine (p : T) : T := top ** p.

  Canonical adjunction_affine : {adj T -> T} :=
    Eval hnf in mk_adjunction affine (is_adjunction_meet I).

  Lemma affine_kernel : is_kernel affine.
  Proof.
    move=> p q. split=> [->|le_ap_q]. exact: meetEr.
    rewrite/affine. apply: meetI=> //. exact: meetEl.
  Qed.

  Lemma affine_strong : is_strong affine.
  Proof.
    move=> p q. by rewrite/affine meetCA.
  Qed.

  Lemma affineI : affine I = I.
  Proof. by rewrite/affine meetxx. Qed.

  Lemma affineP (p q : T) :
    affine p ** q <= q.
  Proof.
    rewrite starMx star1x. exact: meetEl.
  Qed.

  Lemma effineP (p q : T) :
    effine p ** q <= effine q.
  Proof.
    apply: sep_frameR. exact: topI.
  Qed.*)

End QuantaleLaws.

Definition O {T : cll} : T :=
  CLL.zero (CLL.class T).

Definition dual {T : cll} (x : T) : T :=
  x -∘ O.
Notation "x ^⊥" := (dual x) (at level 3, left associativity, format "x ^⊥") : ord_scope.

Definition par {T : cll} (x y : T) : T :=
  (x^⊥ ⊗ y^⊥)^⊥.
Notation "p ⊕ q" := (par p q) (at level 80, right associativity) : ord_scope.

Section CLLLaws.
  Variable T : cll.
  Implicit Types (x y z : T).

  Global Instance dual_proper : Proper (ord_op --> ord_op) (@dual T).
  Proof. move=> x y le_y_x. exact: wand_proper. Qed.

  Lemma dual_involution x :
    x^⊥^⊥ = x.
  Proof.
    apply: le_eq.
    - by case: T x => /=?[?[]].
    - apply/star_adj. exact: starxW.
  Qed.

  Lemma dual_leq x y :
    (x ≤ y) <-> (y^⊥ ≤ x^⊥).
  Proof.
    split=> l.
    - apply/star_adj. rewrite l. exact: starWx.
    - by rewrite -[x]dual_involution -l dual_involution.
  Qed.

  Lemma dual_bot : ⊥^⊥ = ⊤ :> T.
  Proof. exact: wandBx. Qed.

  Lemma dual_top : ⊤^⊥ = ⊥ :> T.
  Proof. by rewrite -[⊥]dual_involution dual_bot. Qed.

  Lemma dual_I : I^⊥ = O :> T.
  Proof. by rewrite/dual wand1x. Qed.

  Lemma dual_O : O^⊥ = I :> T.
  Proof. by rewrite -[I]dual_involution dual_I. Qed.

  Lemma dual_join x y : (x ∨ y)^⊥ = (x^⊥ ∧ y^⊥).
  Proof.
    apply: le_eq.
    - focus; apply/dual_leq; rewrite !dual_involution. exact: joinIl. exact: joinIr.
    - apply/dual_leq. rewrite dual_involution.
      focus; apply/dual_leq; rewrite dual_involution. exact: meetEl. exact: meetEr.
  Qed.

  Lemma dual_meet x y : (x ∧ y)^⊥ = (x^⊥ ∨ y^⊥).
  Proof. by rewrite -[_ ∨ _]dual_involution dual_join !dual_involution. Qed.

  Lemma dual_sup X (F : X -> T) : (sup F)^⊥ = ∀ i, (F i)^⊥.
  Proof.
    apply: le_eq.
    - focus=> i. apply/dual_leq. rewrite !dual_involution. exact: exI.
    - apply/dual_leq. rewrite !dual_involution. focus=> i.
      apply/dual_leq. rewrite dual_involution. exact: allE.
  Qed.

  Lemma dual_inf X (F : X -> T) : (inf F)^⊥ = ∃ i, (F i)^⊥.
  Proof.
    rewrite -[sup _]dual_involution dual_sup.
    do 2 f_equal; fext=> i. by rewrite dual_involution.
  Qed.

  Lemma dual_supc X (P : X -> Prop) (F : X -> T) :
    (∃ i | P i, F i)^⊥ = ∀ i | P i, (F i)^⊥.
  Proof.
    rewrite dual_sup. f_equal; fext=> i. by rewrite dual_sup.
  Qed.

  Lemma dual_infc X (P : X -> Prop) (F : X -> T) :
    (∀ i | P i, F i)^⊥ = ∃ i | P i, (F i)^⊥.
  Proof.
    rewrite dual_inf; f_equal; fext=> i. by rewrite dual_inf.
  Qed.

  Lemma dual_star x y : (x ⊗ y)^⊥ = (x^⊥ ⊕ y^⊥).
  Proof. by rewrite/par !dual_involution. Qed.

  Lemma dual_par x y : (x ⊕ y)^⊥ = (x^⊥ ⊗ y^⊥).
  Proof. by rewrite/par !dual_involution. Qed.

  Lemma cll_wand x y : (x -∘ y) = (x^⊥ ⊕ y).
  Proof.
    apply: le_eq.
    - apply/star_adj. rewrite dual_involution -starA starWx. exact: starxW.
    - apply/star_adj. rewrite/par dual_involution /dual -wandxW starWx.
        by rewrite -{2}[y]dual_involution.
  Qed.

  Lemma dual_wand x y : (x -∘ y)^⊥ = (x ⊗ y^⊥).
  Proof. by rewrite cll_wand dual_par dual_involution. Qed.
End CLLLaws.

Class is_qmorph {X Y : quantale} (f : X -> Y) := mk_is_qmorph {
  radj_of_qcont : Y -> X;
  qcont_is_adjunction :> is_adjunction f radj_of_qcont;
  qcontI : f I = I;
  qcontS : forall x y, f (x ⊗ y) = (f x ⊗ f y)
}.

Class is_cllmorph {X Y : cll} (f : X -> Y) := mk_is_cllmorph {
  is_cllmorph_is_qmorph :> is_qmorph f;
  qcontO : f O = O
}.

Instance cllmorph_is_adjunction {X Y : cll} (f : X -> Y) {H : is_cllmorph f} :
  is_adjunction f radj_of_qcont.
Proof. exact: qcont_is_adjunction. Qed.

Section QuantaleMorphism.
  Variables (X Y : quantale) (f : X -> Y).
  Context {fP : is_qmorph f}.

  Global Instance qcont_proper :
    Proper (ord_op ++> ord_op) f.
  Proof. exact _. Qed.

  Lemma qcontB : f ⊥ = ⊥.
  Proof. apply: adjB. Qed.

  Lemma qcontJ x y : f (x ∨ y) = (f x ∨ f y).
  Proof. apply: adjJ. Qed.

  Lemma qcontE T (F : T -> X) : f (sup F) = ∃ i, f (F i).
  Proof. exact: adjE. Qed.

  Lemma qcontW x y :
    f (x -∘ y) ≤ f x -∘ f y.
  Proof.
    apply/star_adj. by rewrite -qcontS starWx.
  Qed.
End QuantaleMorphism.

Section CLLMorphism.
  Variables (X Y : cll) (f : X -> Y).
  Context {fP : is_qmorph f}.

  Hypothesis qcontD : forall x, f (x^⊥) = (f x)^⊥.

  Lemma qcontP x y : f (x ⊕ y) = (f x ⊕ f y).
  Proof. by rewrite qcontD qcontS !qcontD. Qed.

  Lemma qcontO' : f O = O.
  Proof. by rewrite -dual_I qcontD qcontI dual_I. Qed.

  Lemma qcontW' x y : f (x -∘ y) = (f x -∘ f y).
  Proof. by rewrite cll_wand qcontP qcontD -cll_wand. Qed.

  Lemma qcontT : f ⊤ = ⊤.
  Proof. by rewrite -dual_bot qcontD qcontB dual_bot. Qed.

  (* similarly, preservation of infima... *)
End CLLMorphism.

(** ** The Chu construction *)

Reserved Notation "x ^+" (at level 3, left associativity, format "x ^+").
Reserved Notation "x ^-" (at level 3, left associativity, format "x ^-").

Section Chu.
  Variables (A : quantale).
  Implicit Types (x y z : A).

  Record chu := mk_chu { chuL : A; chuR : A; _ : chuL ⊗ chuR ≤ ⊥ }.
  Arguments mk_chu l r p : rename, clear implicits.
  Implicit Types (p q r : chu).

  Notation "p ^+" := (chuL p).
  Notation "p ^-" := (chuR p).

  Lemma chuP p : p^+ ⊗ p^- ≤ ⊥.
  Proof. by case: p. Qed.

  (*Program Definition dual p := mk_chu (chuR p) (chuL p) _.
  Next Obligation. rewrite starC. exact: chuP. Qed.

  Notation "p ^⊥" := (dual p).*)

  Definition chu_le p q := (chuL p ≤ chuL q) /\ (chuR q ≤ chuR p).

  Lemma chu_le_refl : reflexive chu_le.
  Proof. by split. Qed.

  Lemma chu_le_trans : transitive chu_le.
  Proof. move=> p q r [h1 h2] [h3 h4]. split. by rewrite h1. by rewrite h4. Qed.

  Canonical chu_proMixin := ProMixin chu_le_refl chu_le_trans.
  Canonical chu_proType  := Eval hnf in ProType chu chu_proMixin.

  Lemma chu_eq p q : chuL p = chuL q -> chuR p = chuR q -> p = q.
  Proof.
    move: p q => [l1 r1 h1] [l2 r2 h2] /= e1 e2. destruct e1, e2.
    f_equal. exact: pi.
  Qed.

  Lemma chu_le_eq p q : p ≤ q -> q ≤ p -> p = q.
  Proof.
    move=> [l1 r1] [l2 r2]. apply: chu_eq; exact: le_eq.
  Qed.

  Canonical chu_ordMixin := OrdMixin chu_le_eq.
  Canonical chu_ordType  := Eval hnf in OrdType chu chu_ordMixin.

  (*Lemma le_dual p q :
    (p ≤ q) <-> (q^⊥ ≤ p^⊥).
  Proof. split=> -[]; by split. Qed.*)

  Program Definition chu_inf T (F : T -> chu) := mk_chu (inf (chuL \o F)) (sup (chuR \o F)) _.
  Next Obligation.
    rewrite starxE. focus=> i /=. apply/star_adj.
    apply: allE (i) _ => /=. apply/star_adj. exact: chuP.
  Qed.

  Lemma chu_infP T (F : T -> chu) : is_glb F (chu_inf F).
  Proof.
    move=> x /=. split. move=> [/=h1 h2] i. split. rewrite h1. exact: allE (i) _.
    rewrite -h2. exact: exI (i) _. move=> h. split=> /=. apply: allI => i /=.
    by case: (h i). apply: exE => i /=. by case: (h i).
  Qed.

  Canonical chu_clatMixin := CLatMixin chu_infP.
  Canonical chu_clat := Eval hnf in CLat chu chu_clatMixin.

  Program Definition chu_dual p := mk_chu (chuR p) (chuL p) _.
  Next Obligation. rewrite starC. exact: chuP. Qed.

  Program Definition chu_star p q :=
    mk_chu (p^+ ⊗ q^+) ((p^+ -∘ q^-) ∧ (q^+ -∘ p^-)) _.
  Next Obligation.
    rewrite starxM. apply: meetEl.
    rewrite [p^+ ⊗ _]starC starA starxW. exact: chuP.
  Qed.

  Definition chu_wand p q := chu_dual (chu_star p (chu_dual q)).

  Program Definition chu_I := mk_chu I ⊥ _.
  Next Obligation. by rewrite star1x. Qed.

  Lemma chu_starC p q : chu_star p q ≤ chu_star q p.
  Proof. split=> /=. by rewrite starC. by rewrite meetC. Qed.

  Lemma chu_star1x p : chu_star chu_I p = p.
  Proof.
    apply: le_eq; split=> /=.
    - by rewrite star1x.
    - rewrite wand1x. focus=> //. apply/star_adj.
      rewrite starC. exact: chuP.
    - by rewrite star1x.
    - rewrite wand1x. exact: meetEl.
  Qed.

  Lemma chu_starA p q r :
    chu_star (chu_star p q) r ≤ chu_star p (chu_star q r).
  Proof.
    split=> /=. by rewrite starA. apply: meetI.
    apply: meetEl. rewrite wandxM. apply: meetEl. by rewrite wandxW.
    rewrite !wandxM. apply: meetI. apply: meetEl. apply: meetEr.
    by rewrite !wandxW starC. apply: meetEr. by rewrite wandxW starC.
  Qed.

  Lemma chu_star_adj p q r :
    (chu_star p q ≤ r) <-> (p ≤ chu_wand q r).
  Proof.
    split=> -[/= h1 h2]; split=> /=.
    - apply: meetI. exact/star_adj. apply/star_adj. rewrite h2.
      rewrite starxM. apply: meetEl. by rewrite starxW.
    - rewrite h2. rewrite starxM. apply: meetEr. by rewrite starxW.
    - rewrite h1. rewrite starMx. apply: meetEl. by rewrite starWx.
    - apply: meetI; apply/star_adj; last first. by rewrite starC.
      rewrite h1 starxM. apply: meetEr. by rewrite starxW.
  Qed.

  Canonical chu_quantaleMixin := QuantaleMixin chu_star_adj chu_starC chu_starA chu_star1x.
  Canonical chu_quantale := Eval hnf in Quantale chu chu_quantaleMixin.

  Definition chu_O := chu_dual I.

  Lemma chu_dualP p :
    (p -∘ chu_O) = chu_dual p.
  Proof.
    apply: chu_eq => /=; last first. by rewrite starx1.
    rewrite wand1x. apply: le_eq. exact: meetEr.
    focus. apply/star_adj. rewrite starC. exact: chuP.
  Qed.

  Lemma chu_classical p :
    ((p -∘ chu_O) -∘ chu_O) ≤ p.
  Proof.
    rewrite !chu_dualP. case: p => /=l r o. by split.
  Qed.

  Canonical chu_cllMixin := CLLMixin chu_classical.
  Canonical chu_cll := Eval hnf in CLL chu chu_cllMixin.

  Lemma chu_dualE p : p^⊥ = chu_dual p.
  Proof. exact: chu_dualP. Qed.

  Lemma chu_dual_pos p : p^⊥^+ = p^-.
  Proof. by rewrite chu_dualE. Qed.

  Lemma chu_dual_neg p : p^⊥^- = p^+.
  Proof. by rewrite chu_dualE. Qed.

  Lemma chu_topEp : ⊤^+ = ⊤.
  Proof. apply: top_eq => /=; by focus. Qed.

  Lemma chu_topEn : ⊤^- = ⊥.
  Proof. apply: bot_eq => /=; by focus. Qed.

  Lemma chu_botEp : ⊥^+ = ⊥.
  Proof. by rewrite -dual_top chu_dual_pos chu_topEn. Qed.

  Lemma chu_botEn : ⊥^- = ⊤.
  Proof. by rewrite -dual_top chu_dual_neg chu_topEp. Qed.

  Lemma chu_meetEp (x y : chu) :
    (x ∧ y)^+ = (x^+ ∧ y^+).
  Proof.
    apply: le_eq. focus. exact: allE true _. exact: allE false _.
    cbn. focus=> -[] /=. exact: meetEl. exact: meetEr.
  Qed.

  Lemma chu_meetEn (x y : chu) :
    (x ∧ y)^- = (x^- ∨ y^-).
  Proof.
    apply: le_eq => /=; focus. case=>/=. exact: joinIl. exact: joinIr.
    exact: exI true _. exact: exI false _.
  Qed.

  Lemma chu_joinEp (x y : chu) :
    (x ∨ y)^+ = (x^+ ∨ y^+).
  Proof.
    rewrite -{1}[x]dual_involution -{1}[y]dual_involution -dual_meet.
      by rewrite chu_dual_pos chu_meetEn !chu_dual_neg.
  Qed.

  Lemma chu_joinEn (x y : chu) :
    (x ∨ y)^- = (x^- ∧ y^-).
  Proof.
    rewrite -{1}[x]dual_involution -{1}[y]dual_involution -dual_meet.
      by rewrite chu_dual_neg chu_meetEp !chu_dual_pos.
  Qed.

  Lemma chu_infEp I (F : I -> chu) :
    (inf F)^+ = ∀ i, (F i)^+.
  Proof. by []. Qed.

  Lemma chu_infEn I (F : I -> chu) :
    (inf F)^- = ∃ i, (F i)^-.
  Proof. by []. Qed.

  Lemma dual_supE {C : cll} T (F : T -> C) :
    sup F = (∀ i, (F i)^⊥)^⊥.
  Proof.
    rewrite dual_inf. f_equal; fext=> i. by rewrite dual_involution.
  Qed.

  Lemma chu_supEp T (F : T -> chu) :
    (sup F)^+ = (∃ i, (F i)^+).
  Proof.
    rewrite dual_supE chu_dual_pos chu_infEn.
      by setoid_rewrite chu_dual_neg.
  Qed.

  Lemma chu_supEn T (F : T -> chu) :
    (sup F)^- = (∀ i, (F i)^-).
  Proof.
    rewrite dual_supE chu_dual_neg chu_infEp.
      f_equal; fext=> i. by rewrite chu_dual_pos.
  Qed.

  Lemma chu_starEp p q :
    (p ⊗ q)^+ = (p^+ ⊗ q^+).
  Proof. by []. Qed.

  Lemma chu_starEn p q :
    (p ⊗ q)^- = ((p^+ -∘ q^-) ∧ (q^+ -∘ p^-)).
  Proof. by []. Qed.

  Lemma chu_wandEp p q :
    (p -∘ q)^+ = ((p^+ -∘ q^+) ∧ (q^- -∘ p^-)).
  Proof. by []. Qed.

  Lemma chu_wandEn p q :
    (p -∘ q)^- = (p^+ ⊗ q^-).
  Proof. by []. Qed.

  Lemma chu_IEp :
    I^+ = I.
  Proof. by []. Qed.

  Lemma chu_IEn :
    I^- = ⊥.
  Proof. by []. Qed.

  Lemma chu_OEp : O^+ = ⊥.
  Proof. by []. Qed.

  Lemma chu_OEn : O^- = I.
  Proof. by []. Qed.

  Lemma chu_parEp p q :
    (p ⊕ q)^+ = ((p^- -∘ q^+) ∧ (q^- -∘ p^+)).
  Proof.
    by rewrite/par chu_dual_pos chu_starEn !chu_dual_pos !chu_dual_neg.
  Qed.

  Lemma chu_parEn p q :
    (p ⊕ q)^- = (p^- ⊗ q^-).
  Proof.
    by rewrite/par chu_dual_neg chu_starEp !chu_dual_pos.
  Qed.

  Program Definition chu_embed x : chu :=
    mk_chu x (x -∘ ⊥) _.
  Next Obligation. exact: starxW. Qed.

  Lemma chu_box_def p :
    box p = chu_embed (box p^+).
  Proof.
    apply: le_eq.
    - rewrite{1}/box. focus=> /=i [[[/= aff1 aff2] [/= dup1 dup2]] [le_i_p le_p_i]]. split=> /=.
      + apply: exIc (i^+) _ le_refl. split. split. exact: aff1. exact: dup1. exact: le_i_p.
      + rewrite -dup2 meetxx. apply/star_adj. rewrite aff2.
        suff: i^+ ≤ box p^+. move->. exact: starWx.
        exact: exIc (i^+) _ le_refl.
    - apply: exIc (chu_embed (box p^+)) _ le_refl. split; [split|].
      + split=> /=. exact: is_persistent_is_affine. apply/star_adj.
          by rewrite starBx.
      + split=> /=. by rewrite -dup. rewrite meetxx. by rewrite wandxW -dup.
      + split=> /=. exact: kernel_dec. apply/star_adj. rewrite kernel_dec starC.
        exact: chuP.
  Qed.
  
  Lemma chu_boxEp p : (box p)^+ = box (p^+).
  Proof. by rewrite chu_box_def. Qed.

  Lemma chu_boxEn p : (box p)^- = (box (p^+) -∘ ⊥).
  Proof. by rewrite chu_box_def. Qed.

  Definition why p := dual (box (dual p)).

  Lemma chu_whyEp p : (why p)^+ = (box p^- -∘ ⊥).
  Proof. by rewrite/why chu_dual_pos chu_boxEn chu_dual_pos. Qed.

  Lemma chu_whyEn p : (why p)^- = box p^-.
  Proof. by rewrite/why chu_dual_neg chu_boxEp chu_dual_pos. Qed.

  Lemma why_join p q : why (p ∨ q) = (why p ⊕ why q).
  Proof.
    by rewrite{1}/why dual_join box_star dual_star.
  Qed.

  Lemma why_bot : why ⊥ = O.
  Proof. by rewrite/why dual_bot box_top dual_I. Qed.

  Lemma whyE p q :
    O ≤ q -> q ⊕ q ≤ q -> p ≤ q ->
    why p ≤ q.
  Proof.
    move=> h1 h2 h3. apply/dual_leq. rewrite/why dual_involution. apply: boxI.
    - apply/dual_leq. rewrite dual_involution dual_I. exact: h1.
    - apply/dual_leq. rewrite dual_involution. exact: h2.
    - apply/dual_leq. rewrite !dual_involution. exact: h3.
  Qed.

  (*Lemma chu_embed_wand x y :
    (chu_embed x -∘ chu_embed y)^- ≤ (chu_embed (x -∘ y))^-.
  Proof.
    cbn. by rewrite -star_adj [x ⊗ _]starC starA starxW starWx.
  Qed.

  Lemma chu_embed_wand x y :
    (chu_embed x -∘ chu_embed y)^+ = (x -∘ y).
  Proof.
    apply: le_eq => /=.
    - exact: meetEl.
    - focus. by rewrite -!star_adj starC -starA starxW starxW.
  Qed.*)
End Chu.

Module RA.
  Record mixin_of (T : proType) := Mixin {
    one : T;
    mul : T -> T -> T;
    _ : forall x y, mul x y ≤ mul y x;
    _ : forall x y z, mul (mul x y) z ≤ mul x (mul y z);
    _ : forall x, monotone (mul x);
    _ : forall x, mul one x ≤ x;
    _ : forall x, x ≤ mul one x
  }.

  Section ClassDef.
    Record class_of (T : Type) := Class {
      base : Pro.class_of T;
      mixin : mixin_of (Pro.Pack base T)
    }.
    Local Coercion base : class_of >-> Pro.class_of.

    Structure type := Pack {sort; _ : class_of sort; _ : Type }.
    Local Coercion sort : type >-> Sortclass.

    Variable (T : Type) (cT : type).
    Definition class := let: Pack _ c _ as cT' := cT return class_of cT' in c.
    Definition clone c of phant_id class c := @Pack T c T.
    Let xT := let: Pack T _ _ := cT in T.
    Notation xclass := (class : class_of xT).

    Definition pack b0 (m0 : mixin_of (@Pro.Pack T b0 T)) :=
      fun bT b & phant_id (Pro.class bT) b =>
      fun    m & phant_id m0 m => Pack (@Class T b m) T.

    Definition proType := @Pro.Pack cT xclass xT.
  End ClassDef.

  Module Exports.
    Coercion base : class_of >-> Pro.class_of.
    Coercion mixin : class_of >-> mixin_of.
    Coercion sort : type >-> Sortclass.

    Coercion proType : type >-> Pro.type.
    Canonical proType.

    Notation ra := type.
    Notation RaMixin := Mixin.
    Notation Ra T m := (@pack T _ m _ _ id _ id).

    Notation "[ 'raMixin' 'of' T ]" := (class _ : mixin_of T)
      (at level 0, format "[ 'raMixin'  'of'  T ]") : form_scope.
    Notation "[ 'ra' 'of' T 'for' C ]" := (@clone T C _ idfun)
      (at level 0, format "[ 'ra'  'of'  T  'for'  C ]") : form_scope.
    Notation "[ 'ra' 'of' T ]" := (@clone T _ _ id)
      (at level 0, format "[ 'ra'  'of'  T ]") : form_scope.
  End Exports.
End RA.
Export RA.Exports.

(** ** Operations *)

Definition rmul {T : ra} : T -> T -> T :=
  RA.mul (RA.class T).

Definition rone {T : ra} : T :=
  RA.one (RA.class T).

Infix "*" := rmul : ra_scope.
Notation "1" := rone : ra_scope.
Open Scope ra_scope.

Lemma rmulC {T : ra} (x y : T) : x * y ≤ y * x.
Proof. by case: T x y => T [b []]. Qed.

Lemma rmulA {T : ra} (x y z : T) : (x * y) * z ≤ x * (y * z).
Proof. by case: T x y z => T [b []]. Qed.

Lemma rmulR {T : ra} (x y z : T) : y ≤ z -> x * y ≤ x * z.
Proof. by case: T x y z => T [b []]. Qed.

Lemma rmulL {T : ra} (x y z : T) : x ≤ y -> x * z ≤ y * z.
Proof. move=> le_x_y. by rewrite rmulC (rmulR _ le_x_y) rmulC. Qed.

Instance rmul_proper {T : ra} :
  Proper (ord_op ++> ord_op ++> ord_op) (@rmul T).
Proof.
  move=> x1 x2 lex y1 y2 ley. rewrite rmulR // rmulL //.
Qed.

Lemma rmulAr {T : ra} (x y z : T) : x * (y * z) ≤ (x * y) * z.
Proof.
  rewrite [_ _ z]rmulC [_ x _]rmulC rmulA.
    by rewrite [_ y _]rmulC [_ z _]rmulC.
Qed.

Lemma rmul1l {T : ra} (x : T) : 1 * x ≤ x.
Proof. by case: T x => ?[?[]]. Qed.

Lemma rmul1r {T : ra} (x : T) : x ≤ 1 * x.
Proof. by case: T x => ?[?[]]. Qed.

(** ** Presheaf Quantale *)

Section PresheafQuantale.
  (** The Kripke monoid construction *)
  Variable (X : ra).
  Implicit Types (f g h : presheaf X).

  Program Definition mfun_star f g : presheaf X :=
    mk_mfun (fun (z : X^r) => ∃ x : X, ∃ y : X | (z ≤ x * y :> X), f x ∧ g y) _.
  Next Obligation.
    focus=> l r /= le_x_lr. apply: exI (l) _. apply: exIc (r) _ le_refl.
    rewrite -le_x_lr. exact: H.
  Qed.

  Program Definition mfun_wand f g : presheaf X :=
    mk_mfun (fun (y : X^r) => ∀ x : X, f x → g (x * y)) _.
  Next Obligation.
    focus=> z. apply: allE (z) _. apply/impP. rewrite impEr.
    apply: presheafP. exact: rmulR.
  Qed.

  Definition mfun_one : presheaf X := yo 1.

  Lemma mfun_starC f g : mfun_star f g ≤ mfun_star g f.
  Proof.
    move=> z /=. focus=> x y le. apply: exI (y) _. apply: exIc (x) _ _.
      by rewrite le rmulC. by rewrite meetC.
  Qed.

  Lemma mfun_starA f g h :
    mfun_star (mfun_star f g) h ≤ mfun_star f (mfun_star g h).
  Proof.
    move=> z /=. focus=> x1 x2 le1. apply/impP. focus=> x3 x4 le2.
    rewrite -impP. apply: exI (x3) _. apply: exIc (x4 * x2) _ _.
    by rewrite le1 le2 rmulA. rewrite -meetA. apply: mono.
    apply: exI (x4) _. exact: exIc.
  Qed.

  Lemma mfun_star_adj f g h :
    (mfun_star f g ≤ h) <-> (f ≤ mfun_wand g h).
  Proof.
    split=> le.
    - move=> x /=. focus=> y. rewrite -impP -(le _).
      apply: exI (x) _. apply: exIc (y) _ _ => //.
      by rewrite rmulC.
    - move=> x /=. focus=> y1 y2 le2. rewrite (le _) impP.
      apply: allE (y2) _. apply/impP. rewrite impEr. apply: presheafP.
        by rewrite -rmulC.
  Qed.

  Lemma mfun_oneP f :
    mfun_star mfun_one f = f.
  Proof.
    apply: le_eq.
    - rewrite mfun_star_adj. rewrite presheaf_leE => y /= le_y_o.
      hnf=> x. rewrite prop_impE => fx. apply: presheafP fx.
      by rewrite le_y_o rmulC rmul1l.
    - rewrite presheaf_leE => x fx /=. rewrite prop_exE. exists 1.
      rewrite prop_exEc. exists x. exact: rmul1r. by rewrite prop_meetE.
  Qed.

  Canonical presheaf_quantaleMixin := QuantaleMixin mfun_star_adj mfun_starC mfun_starA mfun_oneP.
  Canonical presheaf_quantale := Eval hnf in Quantale (presheaf X) presheaf_quantaleMixin.

  Lemma presheaf_starE f g (z : X) :
    (f ⊗ g) z = exists x, exists y, (z ≤ x * y) /\ f x /\ g y.
  Proof.
    unfold star => /=. rewrite prop_exE. f_equal; fext=> x.
    rewrite prop_exE. f_equal; fext=> y. rewrite/supguard prop_exE.
    rewrite prop_meetE. apply: pext; by firstorder.
  Qed.

  Lemma presheaf_wandE f g (x : X) :
    (f -∘ g) x = forall y, f y -> g (y * x).
  Proof.
    rewrite/= prop_allE. f_equal; fext=> y. by rewrite prop_impE.
  Qed.

  Lemma yo_star (x y : X) :
    yo (x * y) = (yo x ⊗ yo y).
  Proof.
    apply: le_eq.
    - rewrite yoneda. rewrite presheaf_starE. exists x, y.
      split=> //; split; exact: le_refl.
    - rewrite (@star_adj presheaf_quantale). rewrite yoneda presheaf_wandE => z /= le_z_y.
        by rewrite rmulC le_z_y.
  Qed.

  Lemma yo_one :
    yo 1 = I.
  Proof. by []. Qed.
End PresheafQuantale.

(** ** Ressource Sites *)

Structure rbasis (X : ra) := mk_rbasis {
  rbasis_cov :> X -> Cover X -> Prop;
  _ : forall U x y, rbasis_cov x U -> U y -> y ≤ x;
  _ : forall (U : Cover X) (x y : X), rbasis_cov x U ->
        exists2 C : Pred X, rbasis_cov (y * x) C &
          forall x, C x -> exists2 z, U z & x ≤ y * z
}.

Lemma rbasis_le {X : ra} (U : rbasis X) (C : Cover X) (x y : X) :
  U x C -> C y -> y ≤ x.
Proof. case: U => U le /= _. exact: le. Qed.

Lemma rbasis_rmul {X : ra} (U : rbasis X) (C : Cover X) (x y : X) :
  U x C -> exists2 C', U (y * x) C' &
    forall x, C' x -> exists2 z, C z & x ≤ y * z.
Proof. case: U => U /= _ h. exact: h. Qed.

Section SheafQuantale.
  Variable (X : ra) (U : rbasis X).
  Implicit Types (f g h : sheaf U).

  Definition sheaf_star f g : sheaf U :=
    sheafify U ((f : presheaf X) ⊗ g).

  Program Definition sheaf_wand_S (f : presheaf X) (g : sheaf U) : sheaf U :=
    @mk_sheaf X U (f -∘ g) _.
  Next Obligation.
    move=> C x c le. rewrite presheaf_wandE => /= y fy.
    case: (rbasis_rmul y c) => C' uc H. apply: sheafP uc _ => z cz.
    case: (H z cz) => w cw le'. apply: presheafP le' _.
    move: (le _ cw). rewrite presheaf_wandE. exact.
  Qed.

  Program Definition sheaf_wand f g : sheaf U :=
    sheaf_wand_S f g.

  Definition sheaf_one : sheaf U :=
    sheafify U (I : presheaf X).

  Lemma sheaf_starC f g : sheaf_star f g ≤ sheaf_star g f.
  Proof. rewrite/sheaf_star. apply: mono. by rewrite starC. Qed.

  Lemma sheaf_star_adj f g h :
    (sheaf_star f g ≤ h) <-> (f ≤ sheaf_wand g h).
  Proof.
    split=> le.
    - change (f ≤ (g : presheaf X) -∘ h :> presheaf X).
      rewrite -(@star_adj (presheaf_quantale X)) presheaf_leE => x px.
      apply: (le x). exact: sheafify_inc.
    - apply: allEc (h) _ le_refl. apply: (@sep_cut (presheaf_quantale X)) => //. exact: le.
  Qed.

  Lemma sheaf_starA f g h :
    sheaf_star (sheaf_star f g) h ≤ sheaf_star f (sheaf_star g h).
  Proof.
    apply/sheaf_star_adj. apply: allEc _ le_refl.
    rewrite -(@star_adj (presheaf_quantale X)) starA.
    rewrite/sheaf_star. by rewrite -!sheafify_inc.
  Qed.

  Lemma sheafify_star_strong (f g : presheaf X) :
    f ⊗ sheafify U g ≤ sheafify U (f ⊗ g).
  Proof.
    rewrite starC. apply/star_adj. apply: allEc (sheaf_wand_S f (sheafify U (f ⊗ g))) _ _ => //.
    apply/(@star_adj (presheaf_quantale X)). rewrite starC. exact: sheafify_inc.
  Qed.

  Lemma sheaf_oneP f :
    sheaf_star sheaf_one f = f.
  Proof.
    apply: le_eq.
    - rewrite sheaf_starC /sheaf_one/sheaf_star. rewrite sheafify_star_strong.
        by rewrite starx1 !sheafify_idem.
    - rewrite/sheaf_star. rewrite -sheafify_inc star1x. exact: sheafify_inc.
  Qed.

  Canonical sheaf_quantaleMixin := QuantaleMixin sheaf_star_adj sheaf_starC sheaf_starA sheaf_oneP.
  Canonical sheaf_quantale := Eval hnf in Quantale (sheaf U) sheaf_quantaleMixin.

  Lemma sheafify_star (f g : presheaf X) :
    sheafify U (f ⊗ g) = (sheafify U f ⊗ sheafify U g).
  Proof.
    apply: le_eq. apply: mono. apply: star_proper; exact: sheafify_inc.
    apply/sheafify_adj. rewrite sheafify_star_strong starC.
      by rewrite sheafify_star_strong sheafify_idem starC.
  Qed.

  Lemma embed_star (x y : X) :
    embed U (x * y) = (embed U x ⊗ embed U y).
  Proof.
    by rewrite{1}/embed yo_star sheafify_star.
  Qed.

  Lemma embed_one :
    embed U (1 : X) = I.
  Proof. by []. Qed.

  (* Conjecture: the quantale of ideals on a ressource site
     satisfies a universal property. It is the free quantale
     taking multiplication to star and covers to joins. *)
End SheafQuantale.
 