(** * Basic Category Theory

  Categories with finite products and presheafs.

  There are many "unnecessary" universe annotations in the code below. Remove
  them at your peril, since are just as many bugs in Coq's universe handling
  waiting to be triggered by this code. *)
Require Import base data.fintype.
Set Universe Polymorphism.
Set Printing Universes.
Unset Universe Minimization ToSet.

(** ** Categories *)

Module Cat.
  Polymorphic Cumulative Record mixin_of@{i j} (T : Type@{i}) (hom : T -> T -> Type@{j}) := Mixin {
    id : forall a, hom a a;
    comp : forall a b c, hom a b -> hom b c -> hom a c;
    _ : forall a b (f : hom a b), comp (id a) f = f;
    _ : forall a b (f : hom a b), comp f (id b) = f;
    _ : forall a b c d (f : hom a b) (g : hom b c) (h : hom c d),
        comp (comp f g) h = comp f (comp g h)
  }.
  Notation class_of := mixin_of.

  Section ClassDef.
    Polymorphic Cumulative Structure type@{i j} := Pack
      { sort : Type@{i}; hom : sort -> sort -> Type@{j};
        _ : @mixin_of@{i j} sort hom; sort' : Type@{i}; _ : sort' -> sort' -> Type@{j} }.
    Local Coercion sort : type >-> Sortclass.
    Local Coercion hom  : type >-> Funclass.

    Polymorphic Universe i j.
    Variable (T : Type@{i}) (hom : T -> T -> Type@{j}) (cT : type@{i j}).
    Definition class : mixin_of@{i j} cT := let: Pack _ _ c _ _ := cT return @class_of cT cT in c.

    Definition pack c := @Pack T hom c T hom.
    Definition clone c of phant_id class c := @Pack T hom c T hom.
  End ClassDef.

  Module Exports.
    Coercion sort : type >-> Sortclass.
    Coercion hom  : type >-> Funclass.

    Notation cat := type.
    Notation CatMixin := Mixin.
    Notation Cat T hom m := (@pack T hom m).

    Notation "[ 'catMixin' 'of' hom ]" := (class _ : @mixin_of _ hom)
      (at level 0, format "[ 'catMixin'  'of'  hom ]") : form_scope.
    Notation "[ 'cat' 'of' hom 'for' C ]" := (@clone _ hom C _ idfun)
      (at level 0, format "[ 'cat'  'of'  hom  'for'  C ]") : form_scope.
    Notation "[ 'cat' 'of' hom ]" := (@clone _ hom _ _ (fun x => x))
      (at level 0, format "[ 'cat'  'of'  hom ]") : form_scope.
  End Exports.
End Cat.
Export Cat.Exports.

(** *** Operations and laws *)

Definition hom@{i j} {T : cat@{i j}} (a b : T) : Type@{j} := Cat.hom a b.
Arguments hom {T} a b.

Definition cid@{i j} {T : cat@{i j}} {a : T} : hom@{i j} a a :=
  Cat.id@{i j} (Cat.class@{i j} T) a.

Definition comp@{i j} {T : cat@{i j}} : forall a b c, hom a b -> hom b c -> hom a c :=
  @Cat.comp _ _ (Cat.class T).
Arguments comp {T a b c} f g.
Notation "f ∘ g" := (comp f g) : cat_scope.
Open Scope cat_scope.

Section Lemmas.
  Context {C : cat}.
  Implicit Types (a b c d : C).

  Lemma idL a b (f : hom a b) : cid ∘ f = f.
  Proof. by case: C a b f => /=T hom[]. Qed.

  Lemma idR a b (f : hom a b) : f ∘ cid = f.
  Proof. by case: C a b f => /=T hom[]. Qed.

  Lemma compA a b c d (f : hom a b) (g : hom b c) (h : hom c d) :
    (f ∘ g) ∘ h = f ∘ (g ∘ h).
  Proof. by case: C a b c d f g h => /=T hom[]. Qed.
End Lemmas.

Hint Rewrite @idL @idR @compA : cat.
Ltac csimpl := repeat first [autorewrite with cat|move=>/=].

(** ** Categories with finite products *)

Module CartCat.
  Polymorphic Cumulative Record mixin_of@{i j}
    (T : Type@{i}) (hom : T -> T -> Type@{j}) (m : Cat.mixin_of@{i j} hom) := Mixin {
    unit : T;
    bang : forall a : T, hom a unit;
    prod : T -> T -> (T : (Type@{i}));
    pi1 : forall a b : T, hom (prod a b) a : (Type@{j});
    pi2 : forall a b : T, hom (prod a b) b : (Type@{j});
    pair : forall a b c : T, (hom a b : (Type@{j})) -> (hom a c : (Type@{j})) -> (hom a (prod b c) : (Type@{j}));
    _ : forall (a b c : T) (f : hom a b) (g : hom a c),
      Cat.comp m (pair f g) (pi1 b c) = f;
    _ : forall (a b c : T) (f : hom a b) (g : hom a c),
      Cat.comp m (pair f g) (pi2 b c) = g;
    _ : forall (a b c : T) (f : hom a (prod b c)),
      pair (Cat.comp m f (pi1 b c)) (Cat.comp m f (pi2 b c)) = f;
    _ : forall (a : T) (f : hom a unit), f = bang a
  }.

  Section ClassDef.
    Polymorphic Cumulative Record class_of@{i j} (T : Type@{i}) (hom : T -> T -> Type@{j}) := Class {
      base : @Cat.mixin_of T hom;
      mixin : @mixin_of T hom base
    }.
    Local Coercion base : class_of >-> Cat.mixin_of.

    Polymorphic Cumulative Structure type@{i j} := Pack {sort : Type@{i}; hom : sort -> sort -> Type@{j};
      _ : @class_of@{i j} sort hom; sort' : Type@{i}; _ : sort' -> sort' -> Type@{j} }.
    Local Coercion sort : type >-> Sortclass.
    Local Coercion hom : type >-> Funclass.

    Polymorphic Universes i j.
    Variable (T : Type@{i}) (hom : T -> T -> Type@{j}) (cT : type@{i j}).
    Definition class : class_of@{i j} cT := let: Pack _ _ c _ _ as cT' := cT return class_of cT' in c.
    Definition clone c of phant_id class c := @Pack T hom c T hom.
    Let xT : Type@{i} := let: Pack T _ _ _ _ := cT in T.
    Let xH : xT -> xT -> Type@{j} := let: Pack _ H _ _ _ := cT return cT -> cT -> Type@{j} in H.
    Notation xclass := (class : @class_of xT xH).

    Definition pack (T : Type@{i}) (hom : T -> T -> Type@{j})
               (mc : @Cat.mixin_of@{i j} T hom) (mp : @mixin_of@{i j} T hom mc) : type@{i j} :=
      @Pack T hom (@Class T hom mc mp) T hom.

    Definition cat@{i j} := @Cat.Pack@{i j} cT cT xclass xT xH.
  End ClassDef.

  Module Exports.
    Coercion base : class_of >-> Cat.mixin_of.
    Coercion mixin : class_of >-> mixin_of.
    Coercion sort : type >-> Sortclass.
    Coercion hom  : type >-> Funclass.
    Coercion cat  : type >-> Cat.type.
    Canonical cat.

    Notation cartcat := type.
    Notation CartCatMixin := Mixin.
    Notation CartCat T hom mc mp := (@pack T hom mc mp).
  End Exports.
End CartCat.
Export CartCat.Exports.

(** *** Operations and laws *)

Definition cunit {T : cartcat} : T := CartCat.unit (CartCat.class T).
Definition cbang {T : cartcat} (a : T) : T a cunit :=
  CartCat.bang (CartCat.class T) a.
Definition cprod {T : cartcat} (a b : T) : T :=
  CartCat.prod (CartCat.class T) a b.
Definition cproj1 {T : cartcat} (a b : T) : T (cprod a b) a :=
  CartCat.pi1 (CartCat.class T) a b.
Definition cproj2 {T : cartcat} (a b : T) : T (cprod a b) b :=
  CartCat.pi2 (CartCat.class T) a b.
Definition cpair {T : cartcat} {a b c : T} (f : T a b) (g : T a c) : T a (cprod b c) :=
  CartCat.pair (CartCat.class T) f g.

Notation "1" := (@cunit _) : cat_scope.
Notation "!" := (@cbang _ _) : cat_scope.

Notation "a × b" := (@cprod _ a b) (at level 50) : cat_scope.
Notation π₁ := (@cproj1 _ _ _).
Notation π₂ := (@cproj2 _ _ _).
Notation "⟨ f , g ⟩" := (@cpair _ _ _ _ f g)
  (format "⟨ f ,  g ⟩") : cat_scope.

Section CartCatLaws.
  Variable A : cartcat.
  Implicit Types (a b c d : A).

  Lemma unit_eta a (f : hom a 1) : f = !.
  Proof. case: A a f => T hom [base []]//. Qed.

  Lemma pair_pi1 a b c (f : hom a b) (g : hom a c) :
    ⟨f,g⟩ ∘ π₁ = f.
  Proof. case: A a b c f g => ??[?[]]//. Qed.

  Lemma pair_pi2 a b c (f : hom a b) (g : hom a c) :
    ⟨f,g⟩ ∘ π₂ = g.
  Proof. case: A a b c f g => ??[?[]]//. Qed.

  Lemma pair_eta a b c (f : hom a (b × c)) :
    ⟨f∘π₁,f∘π₂⟩ = f.
  Proof. case: A a b c f => ??[?[]]//. Qed.

  Lemma pair_pi1_comp a b c d (f : hom a b) (g : hom a c) (h : hom b d) :
    ⟨f, g⟩ ∘ (π₁ ∘ h) = f ∘ h.
  Proof. by rewrite <-compA, pair_pi1. Qed.

  Lemma pair_pi2_comp a b c d (f : hom a b) (g : hom a c) (h : hom c d) :
    ⟨f, g⟩ ∘ (π₂ ∘ h) = g ∘ h.
  Proof. by rewrite <-compA, pair_pi2. Qed.

  Lemma pair_eta_id a b :
    ⟨π₁,π₂⟩ = cid :> hom (a × b) (a × b).
  Proof. rewrite <-(pair_eta cid). by csimpl. Qed.

  Definition first c a b (f : hom a b) : hom (a × c) (b × c) :=
    ⟨π₁∘f,π₂⟩.

  Definition second a b c (f : hom b c) : hom (a × b) (a × c) :=
    ⟨π₁,π₂∘f⟩.

  Hint Rewrite @pair_pi1 @pair_pi2 @pair_pi1_comp @pair_pi2_comp @pair_eta_id : cat.

  Lemma first_id a b : first b (@cid _ a) = cid.
  Proof. unfold first. by csimpl. Qed.

  Lemma second_id a b : second b (@cid _ a) = cid.
  Proof. unfold second. by csimpl. Qed.

  Lemma first_comp a b c d (f : hom a b) (g : hom b c) :
    first d f ∘ first d g = first d (f ∘ g).
  Proof.
    rewrite <- (pair_eta (first _ f ∘ first _ g)). unfold first. by csimpl.
  Qed.

  Lemma second_comp a b c d (f : hom a b) (g : hom b c) :
    second d f ∘ second d g = second d (f ∘ g).
  Proof.
    rewrite <- (pair_eta (second _ f ∘ second _ g)). unfold second. by csimpl.
  Qed.

  Lemma first_second a b c d (f : hom a b) (g : hom c d) :
    first _ f ∘ second _ g = second _ g ∘ first _ f.
  Proof.
    rewrite <- (pair_eta (first _ f ∘ second _ g)), <- pair_eta.
    unfold first, second. by csimpl.
  Qed.
End CartCatLaws.

Hint Rewrite @pair_pi1 @pair_pi2 @pair_pi1_comp @pair_pi2_comp
     @pair_eta_id @first_id @second_id @first_comp @second_comp : cat.

(** *** Finite products *)

Section FiniteProducts.
  Variable (A : cartcat).

  Fixpoint oprod {n : nat} : (fin n -> A) -> A :=
    match n with
    | 0 => fun=> 1
    | m.+1 => fun (f : option (fin m) -> A) => f None × (oprod (f \o Some))
    end.

  Fixpoint proj {n : nat} : forall {f : fin n -> A} (i : fin n), hom (oprod f) (f i) :=
    match n with
    | 0 => fun _ d => match d with end
    | m.+1 => fun (f : option (fin m) -> A) (d : option (fin m)) =>
      match d with
      | Some i => π₂ ∘ @proj m (Some >> f) i
      | None => π₁
      end
    end.

  Fixpoint hprod {n : nat} : forall {f : fin n -> A} {a : A}, (forall i, hom a (f i)) -> hom a (oprod f) :=
    match n with
    | 0 => fun _ _ _ => !
    | m.+1 => fun (f : option (fin m) -> A) (a : A) (F : forall i, hom a (f i)) =>
                ⟨ F None, @hprod m (Some >> f) a (fun i => F (Some i)) ⟩
    end.

  Lemma proj_hprod {n : nat} {f : fin n -> A} (a : A) (F : forall i : fin n, hom a (f i)) (i : fin n) :
    hprod F ∘ proj i = F i.
  Proof.
    elim: n f F i => //=m ih f F [i|]; csimpl=>//. by apply (ih (Some >> f)).
  Qed.

  Lemma hprod_eta {n : nat} {f : fin n -> A} {a : A} (F : hom a (oprod f)) :
    hprod (fun i => F ∘ proj i) = F.
  Proof.
    elim: n f F => /=[_ F|m ih f F]. by rewrite [F]unit_eta.
    rewrite -{3}[F]pair_eta. f_equal. rewrite <-(ih (Some >> f) (F ∘ π₂)).
    f_equal. fext=>i. by csimpl.
  Qed.
End FiniteProducts.

(** ** Presheafs *)

Polymorphic Cumulative Structure presheaf@{i j k} (A : cat@{i j}) := mk_presheaf
  { omap :> A -> Type@{k}
  ; map : forall (a b : A), A b a -> omap a -> omap b
  ; _ : forall (a : A) (x : omap a), map (cid : A a a) x = x
  ; _ : forall (a b c : A) (f : A a b) (g : A b c) (x : omap c),
      map f (map g x) = map (f ∘ g) x
  }.
Arguments mk_presheaf {A} omap map _ _.

Lemma presheaf_eq (A : cat) (F G : presheaf A) :
  (omap F; @map A F) = (omap G; @map A G) :> { omap : A -> Type & forall a b, A b a -> omap a -> omap b } ->
  F = G.
Proof.
  destruct F, G => /= p. move: p..1 p..2 => /= {p}.
  destruct 0. cbn. destruct 1. f_equal; exact: pi.
Qed.

Section PresheafLemmas.
  Variables (A : cat) (F : presheaf A).

  Lemma map1 a (x : F a) : map (cid : A a a) x = x.
  Proof. by case: F a x. Qed.

  Lemma mapC a b c (f : A a b) (g : A b c) (x : F c) :
    map f (map g x) = map (f ∘ g) x.
  Proof. by case: F a b c f g x. Qed.
End PresheafLemmas.

Hint Rewrite map1 mapC (*mapC_comp*) : cat.

(** ** Natural transformations between presheafs *)
(* Not cumulative, since the type of natural transformations wouldn't be cumulative anyway *)
Polymorphic Structure pnat@{i j k} {A : cat@{i j}} (F G : presheaf@{i j k} A) := mk_pnat
  { components :> forall (a : A), F a -> G a
  ; _ : forall (a b : A) (f : A b a) (x : F a),
      components (map f x) = map f (components x)
  }.
Arguments mk_pnat {A F G} eta _ : rename.

(** *** Laws *)

Lemma pnat_eq (A : cat) (F G : presheaf A) (eta eps : pnat F G) :
  components eta = components eps -> eta = eps.
Proof.
  destruct eta,eps => /= E. destruct E. f_equal. exact: pi.
Qed.

Lemma natP (A : cat) (F G : presheaf A) (eta : pnat F G) (a b : A) (f : hom b a) (x : F a) :
  eta b (map f x) = map f (eta a x).
Proof. by case: eta. Qed.

Hint Rewrite natP : cat.

(** ** Category of presheafs *)

Section PresheafCategory.
  Polymorphic Universes i j k.
  Variables (A : cat@{i j}).
  Implicit Types (F G H : presheaf@{i j k} A).

  Program Definition identity_nat F : pnat@{i j k} F F :=
    mk_pnat (fun _ x => x) (fun _ _ _ _ => erefl).

  Program Definition compose_nat F G H (eta : pnat@{i j k} F G) (eps : pnat@{i j k} G H) : pnat@{i j k} F H :=
    mk_pnat (fun a x => eps a (eta a x)) _.
  Next Obligation. by csimpl. Qed.

  Lemma nat_idL F G (eta : pnat F G) :
    compose_nat (identity_nat F) eta = eta.
  Proof.
    apply: pnat_eq; fext=>/=a. by csimpl.
  Qed.

  Lemma nat_idR F G (eta : pnat F G) :
    compose_nat eta (identity_nat G) = eta.
  Proof.
    apply: pnat_eq; fext=>/=a. by csimpl.
  Qed.

  Lemma nat_compA F G H I (e1 : pnat F G) (e2 : pnat G H) (e3 : pnat H I) :
    compose_nat (compose_nat e1 e2) e3 = compose_nat e1 (compose_nat e2 e3).
  Proof.
    apply: pnat_eq; fext=>/=a. by csimpl.
  Qed.

  Polymorphic Universe u v.
  (*Constraint i <= u. Constraint j <= u. Constraint k < u.
  Constraint i <= v. Constraint j <= v.*)
  Canonical presheaf_catMixin :=
    @CatMixin@{u v} (presheaf@{i j k} A) (pnat@{i j k}) identity_nat compose_nat nat_idL nat_idR nat_compA.
  Canonical PRESHEAF := Eval hnf in Cat (presheaf@{i j k} A) pnat presheaf_catMixin.
End PresheafCategory.

(** *** Terminal object in the category of presheafs *)

Section PresheafTerminal.
  Polymorphic Universes i j k.
  Variable (A : cat@{i j}).
  Implicit Types (F G : presheaf@{i j k} A).

  Program Definition constant_presheaf (B : Type@{j}) : presheaf@{i j k} A :=
    mk_presheaf (fun=> B) (fun _ _ _ => id) _ _.

  Definition presheaf_terminal : presheaf A := constant_presheaf unit.

  Program Definition presheaf_bang (F : presheaf A) : pnat F presheaf_terminal :=
    @mk_pnat A F presheaf_terminal (fun _ _ => tt) _.

  Lemma presheaf_terminal_eta (F : presheaf A) (eta : pnat F presheaf_terminal) :
    eta = presheaf_bang F.
  Proof.
    apply: pnat_eq. fext=> a x /=. by case: (eta a x).
  Qed.
End PresheafTerminal.

(** *** Products in the category of presheafs *)

(*Polymorphic Cumulative Record Pair (A B : Type) : Type :=
  mk_ppair { pfst : A; psnd : B }.*)

Section PresheafBinaryProducts.
  Polymorphic Universes i j k.
  Variable (A : cat@{i j}).
  Implicit Types (F G H : presheaf@{i j k} A).

  (*Program Definition PProd@{i j} F G : presheaf@{i j} A :=
    mk_presheaf (fun a => Pair (F a) (G a)) (fun a b f p => mk_ppair (map@{i j} f (pfst p)) (map@{i j} f (psnd p))) _ _.
  Next Obligation. destruct x as [x y] => /=. by csimpl. Qed.
  Next Obligation. destruct x as [x y] => /=. by csimpl. Qed.*)

  Program Definition PProd@{i j k} F G : presheaf@{i j k} A :=
    mk_presheaf (fun a => (F a * G a)%type) (fun a b f p => (map@{i j k} f p.1, map@{i j k} f p.2)) _ _.
  Next Obligation. move=>/=. by csimpl. Qed.
  Next Obligation. move=>/=. by csimpl. Qed.

  Program Definition nat_pi1 {F G} : pnat@{i j k} (PProd F G) F :=
    mk_pnat (fun a p => p.1) _.

  Program Definition nat_pi2 {F G} : pnat@{i j k} (PProd F G) G :=
    mk_pnat (fun a p => p.2) _.

  Program Definition nat_pair {F G H} (eps : pnat@{i j k} F G) (eta : pnat@{i j k} F H) : pnat@{i j k} F (PProd G H) :=
    mk_pnat (fun a x => (eps a x, eta a x)) _.
  Next Obligation. by csimpl. Qed.

  Lemma nat_pair1 F G H (eps : pnat@{i j k} F G) (eta : pnat@{i j k} F H) :
    nat_pair eps eta ∘ nat_pi1 = eps.
  Proof. exact: pnat_eq. Qed.

  Lemma nat_pair2 F G H (eps : pnat@{i j k} F G) (eta : pnat@{i j k} F H) :
    nat_pair eps eta ∘ nat_pi2 = eta.
  Proof. exact: pnat_eq. Qed.

  Lemma nat_pair_eta F G H (eta : pnat@{i j k} F (PProd@{i j k} G H)) :
    nat_pair (eta ∘ nat_pi1) (eta ∘ nat_pi2) = eta.
  Proof.
    apply: pnat_eq. fext=> a x /=. csimpl. by destruct (eta a x).
  Qed.

  Lemma map_pairE F G (a b : A) (f : A b a) (x : F a) (y : G a) :
    @map A (PProd F G) a b f (x, y) = (map f x, map f y).
  Proof. by []. Qed.
End PresheafBinaryProducts.

(** *** Cartesian category of presheafs *)

Definition presheaf_cartcatMixin@{i j k u v} (A : cat@{i j}) :
  CartCat.mixin_of@{u v} (presheaf_catMixin@{i j k u v} A) :=
    @CartCatMixin _ _ (presheaf_catMixin A)
      (presheaf_terminal A)
      (@presheaf_bang A)
      (@PProd A)
      (@nat_pi1 A)
      (@nat_pi2 A)
      (@nat_pair A)
      (@nat_pair1@{i j k u v} A)
      (@nat_pair2@{i j k u v} A)
      (@nat_pair_eta@{i j k u v} A)
      (@presheaf_terminal_eta A).
Canonical presheaf_cartcatMixin.
Definition presheaf_cartcat@{i j k u v} (A : cat@{i j}) :=
  Eval hnf in @CartCat.pack@{u v}
    (presheaf@{i j k} A)
    (@pnat@{i j k} A)
    (presheaf_catMixin@{i j k u v} A)
    (presheaf_cartcatMixin@{i j k u v} A).
Canonical presheaf_cartcat.

(** ** Yoneda lemma *)

Polymorphic Definition y {A : cat} (a : A) : presheaf A :=
  mk_presheaf (fun (b : A) => A b a) (fun b c f g => (f : A c b) ∘ g)
              (fun _ _ => idL _) (fun _ _ _ _ _ _ => esym (compA _ _ _)).

Section Yoneda.
  Variable A : cat.

  (* The hom-set functor *)

  (*Polymorphic Program Definition y (a : A) : presheaf A :=
    mk_presheaf (fun (b : A) => A b a) (fun b c f g => (f : A c b) ∘ g) (fun _ _ => idL _) (fun _ _ _ _ _ _ => esym (compA _ _ _)).
  Next Obligation. exact: idL. Qed.
  Next Obligation. by csimpl. Qed.*)

  Program Definition ymap (a b : A) (f : A a b) : pnat (y a) (y b) :=
    mk_pnat (fun c g => g ∘ f) _.
  Next Obligation. by csimpl. Qed.

  Lemma ymap1 (a : A) : ymap (@cid A a) = identity_nat (y a).
  Proof. apply: pnat_eq; fext=> b /= f. by csimpl. Qed.

  Lemma ymapC (a b c : A) (f : A a b) (g : A b c) :
    ymap f ∘ ymap g = ymap (f ∘ g).
  Proof. apply: pnat_eq; fext=> d /= h. by csimpl. Qed.

  (*Canonical HOM : functor A (PRESHEAF A) := mk_functor y ymap ymap1 ymapC.*)

  (* The yoneda lemma *)

  Section Lemma.
    Variable (F : presheaf A) (a : A).

    Definition yred (f : pnat (y a) F) : F a := f a cid.

    Program Definition yexp (x : F a) : pnat (y a) F :=
      mk_pnat (fun b (f : A b a) => map f x) _.
    Next Obligation.
      by csimpl.
    Qed.

    Lemma yred_cl (eta : pnat (y a) F) :
      yexp (yred eta) = eta.
    Proof.
      apply: pnat_eq; fext=> /=b f. rewrite -natP /=. by csimpl.
    Qed.

    Lemma yred_cr (x : F a) :
      yred (yexp x) = x.
    Proof. cbn. by csimpl. Qed.

    Canonical yred_is_equiv := mk_is_equiv yexp yred_cl yred_cr.
    Canonical YRED : hom (y a) F <~> F a := mk_equiv yred yred_is_equiv.
  End Lemma.

  (* Naturality of YRED *)

  Lemma yred_nat (F : presheaf A) (a b : A) (f : A a b) (eta : hom (y b) F) :
    yred (ymap f ∘ eta) = map f (yred eta).
  Proof. cbn. unfold yred. csimpl => /=. rewrite -natP. cbn. by csimpl. Qed.

  Lemma yred_natF (F G : presheaf A) (eps : hom F G) (a : A) (eta : hom (y a) F) :
    yred (eta ∘ eps) = eps a (yred eta).
  Proof. by []. Qed.

  (* y is fully faithful *)

  Lemma y_hom_iso (a b : A) :
    hom (y a) (y b) <~> hom a b.
  Proof. exact: YRED. Qed.
End Yoneda.

(** ** Exponentials in presheafs *)

Section PresheafExponentials.
  Polymorphic Universe i j k.
  Variable A : cat@{i j}.
  Implicit Types (F G H : presheaf@{i j k} A).

  Local Obligation Tactic := idtac.

  Definition PExp_obj (F G : presheaf@{i j k} A) (a : A) : Type@{k} :=
    { eta : forall (b : A), A b a -> G b -> F b
    | forall (b c : A) (f : A c b) (g : A b a) (x : G b),
          map f (eta b g x) = eta c (f ∘ g) (map f x) }.

  Definition mk_pexp (F G : presheaf@{i j k} A) (a : A)
             (eta : forall (b : A), A b a -> G b -> F b) :
    (forall (b c : A) (f : A c b) (g : A b a) (x : G b),
        map f (eta b g x) = eta c (f ∘ g) (map f x)) ->
    PExp_obj F G a := exist _ eta.

  Definition PExp_hom (F G : presheaf@{i j k} A) (a b : A) (f : A b a) (H : PExp_obj F G a) :
    PExp_obj F G b.
    refine (exist _ _ _). Unshelve. Focus 2. intros c g x. apply H.1. exact (g ∘ f). exact x.
    abstract (cbn; intros c d g h x; destruct H as [H HP]; cbn; rewrite HP; by rewrite->compA).
  Defined.

  Program Definition PExp (F G : presheaf@{i j k} A) : presheaf@{i j k} A :=
    mk_presheaf (PExp_obj F G) (@PExp_hom F G) _ _.
  Next Obligation.
    move=> F G a [H HP] /=. unfold PExp_hom. cbn. apply: sig_eq => /=.
    fext=> c g x. by csimpl.
  Qed.
  Next Obligation.
    move=> F G a b c f g [H HP]. apply: sig_eq => /=. fext=> d x h. by csimpl.
  Qed.

  (*Definition PExp F G : presheaf A.
    refine (mk_presheaf (fun (a : A) => pnat ((y a : presheaf_cartcat A) × G) F)
                (fun (a b : A) (f : A b a) (F : pnat ((y a : presheaf_cartcat A) × G) F) =>
                   first G (ymap f) ∘ F) _ _).*)
  (*Program Definition PExp (F G : presheaf@{i j k} A) : presheaf@{i j k} A :=
    mk_presheaf (fun (a : A) => pnat@{i j k} (y@{i j} a × G) F)
                (fun (a b : A) (f : A b a) (F : pnat@{i j k} (y a × G) F) =>
                   first G (ymap f) ∘ F) _ _.
  Next Obligation.
    move=> /= F G a f. rewrite-> ymap1, first_id. by csimpl.
  Qed.
  Next Obligation.
    move=> /= F G a b c f g h. rewrite <-compA. rewrite-> first_comp.
      f_equal. f_equal. exact: ymapC.
  Qed.*)

  Program Definition nat_apply F G : pnat@{i j k} (PExp F G × G) F :=
    mk_pnat (fun a etap => etap.1 a cid etap.2) _.
  Next Obligation.
    move=> /= F G a b f [eta x]/=. csimpl. rewrite (svalP eta). by csimpl.
  Qed.

  Program Definition nat_lam F G H (f : pnat@{i j k} (F × G) H) : pnat@{i j k} F (PExp H G) :=
    mk_pnat (fun a x => @mk_pexp H G a (fun b h y => f b (map h x, y)) _) _.
  Next Obligation.
    move=> F G H f a x b c g h y /=. rewrite -natP -mapC//.
  Qed.
  Next Obligation.
    move=> /= F G H f a b g x. apply: sig_eq => /=. fext=>/=c h y. by csimpl.
  Qed.

  Lemma nat_lam_beta F G H (f : pnat (F × G) H) :
    first G (nat_lam f) ∘ nat_apply H G = f.
  Proof.
    apply: pnat_eq; fext=> a -[x y] /=. by csimpl.
  Qed.

  Lemma nat_lam_eta F G :
    nat_lam (nat_apply F G) = cid.
  Proof.
    apply: pnat_eq; fext=> a /= eta. apply: sig_eq; fext=> b f y /=. by csimpl=>/=.
  Qed.

  Lemma nat_lam_nat F G H I (f : hom F G) (g : hom (G × H) I) :
    nat_lam (first H f ∘ g) = f ∘ nat_lam g.
  Proof.
    apply: pnat_eq; fext=> a x. apply: sig_eq; fext=> b h y/=.
      by csimpl=> /=.
  Qed.
End PresheafExponentials.
