Require Import base.

Inductive ext (X : Type) : Type :=
| bound : ext X
| shift : X -> ext X.

Definition scons {X Y} (x : Y) (f : X -> Y) (i : ext X) : Y :=
  if i is shift i then f i else x.

Notation "x .: f" := (@scons _ _ x f) (at level 55).

Definition upr {X Y} (f : X -> Y) : ext X -> ext Y :=
  bound _ .: (f \; @shift _).
  (*if x is shift x then shift (f x) else bound _.*)

(** ** Term structure *)

Inductive tm (X : Type) : Type :=
| var  : X -> tm X
| app  : tm X -> tm X -> tm X
| lam  : tm (ext X) -> tm X
| pi   : tm X -> tm (ext X) -> tm X
| univ : tm X.

Fixpoint map {X Y} (f : X -> Y) (s : tm X) : tm Y :=
  match s with
  | var x => var (f x)
  | app s t => app (map f s) (map f t)
  | lam s => lam (map (upr f) s)
  | pi s t => pi (map f s) (map (upr f) t)
  | univ => univ _
  end.

Definition up {X Y} (f : X -> tm Y) : ext X -> tm (ext Y) :=
  var (bound _) .: (f \; map (@shift _)).
(*Definition up {X Y} (f : X -> tm Y) (x : ext X) : tm (ext Y) :=
  if x is shift x then map (@shift _) (f x) else var (bound _).*)

Fixpoint bind {X Y} (f : X -> tm Y) (s : tm X) : tm Y :=
  match s with
  | var x => f x
  | app s t => app (bind f s) (bind f t)
  | lam s => lam (bind (up f) s)
  | pi s t => pi (bind f s) (bind (up f) t)
  | univ => univ _
  end.

Definition beta {X} (s : tm (ext X)) (t : tm X) : tm X :=
  bind (t .: @var _) s.

Lemma upr_comp {X Y Z} (f : X -> Y) (g : Y -> Z) :
  upr f \; upr g = upr (f \; g).
Proof. by fext=> -[]. Qed.

Lemma map_comp {X Y Z} (f : X -> Y) (g : Y -> Z) (s : tm X) :
  map g (map f s) = map (f \; g) s.
Proof.
  elim: s Y Z f g => //={X}[X s ihs t iht|X s ih|X s ihs t iht] Y Z f g.
  - by rewrite ihs iht.
  - by rewrite ih upr_comp.
  - by rewrite ihs iht upr_comp.
Qed.

(** ** Traditional type system *)

Inductive has_type (X : Type) (Gamma : X -> tm X) : tm X -> tm X -> Prop :=
| of_var (i : X) : has_type Gamma (var i) (Gamma i)
| of_app (s t A : tm X) (B : tm (ext X)) :
    has_type Gamma s (pi A B) ->
    has_type Gamma t A ->
    has_type Gamma (app s t) (beta B t)
| of_lam (A : tm X) (s B : tm (ext X)) :
    @has_type (ext X) (map (@shift _) A .: (Gamma \; map (@shift _))) s B ->
    has_type Gamma (lam s) (pi A B).

Lemma has_type_map (X Y : Type) (Gamma : X -> tm X) (Delta : Y -> tm Y) (f : X -> Y) :
  (forall i : X, map f (Gamma i) = Delta (f i)) ->
  forall s A,
    has_type Gamma s A ->
    has_type Delta (map f s) (map f A).
Proof.
  move=> H s A tp. elim: tp Y Delta f H => {X Gamma s A}.
  - move=> X Gamma i Y Delta f H /=. rewrite H. exact: of_var.
  - move=> X Gamma s t A B _ ih1 _ ih2 Y Delta f H /=.
    have E: map f (beta B t) = beta (map (upr f) B) (map f t) by admit.
    rewrite E. apply: of_app. exact: ih1. exact: ih2.
  - move=> X Gamma A s B _ ih Y Delta f H /=. apply: of_lam.
    apply: ih. case=>[|x]. by rewrite !map_comp.
    rewrite !map_comp /=. change (@shift _ \; upr f) with (f \; @shift _).
    rewrite -map_comp. by rewrite H.
Admitted.

Lemma weakening X (Gamma : X -> tm X) (s A : tm X) (B : tm (ext X)) :
  has_type Gamma s A ->
  has_type (B .: (Gamma \; map (@shift _))) (map (@shift _) s) (map (@shift _) A).
Proof.
  exact: has_type_map.
Qed.

Lemma has_type_bind (X Y : Type) (Gamma : X -> tm X) (Delta : Y -> tm Y) (f : X -> tm Y) :
  (forall i : X, has_type Delta (f i) (bind f (Gamma i))) ->
  forall s A,
    has_type Gamma s A ->
    has_type Delta (bind f s) (bind f A).
Proof.
  move=> H s A tp. elim: tp Y Delta f H => {X Gamma s A}.
  - move=> X Gamma i Y Delta f H. exact: H.
  - move=> X Gamma s t A B _ ih1 _ ih2 Y Delta f H /=.
    have E: bind f (beta B t) = beta (bind (up f) B) (bind f t) by admit.
    rewrite E. apply: of_app. exact: ih1. exact: ih2.
  - move=> X Gamma A s B _ ih Y Delta f H /=. apply: of_lam.
    apply: ih. case=>[|i].
    + cbn. have E: (bind (up f) (map (shift (X:=X)) A)) = map (shift (X:=Y)) (bind f A) by admit.
      rewrite E. exact: of_var.
    + cbn. have E: (bind (up f) (map (shift (X:=X)) (Gamma i))) = map (@shift _) (bind f (Gamma i)) by admit.
      rewrite E. apply: weakening. exact: H.
Admitted.

(** ** Type system *)

Inductive has_type (X Y : Type) (Ctm : Y -> tm X) (Cty : Y -> tm X) : tm X -> tm X -> Prop :=
| of_var (i : Y) : has_type Ctm Cty (Ctm i) (Cty i)
| of_app (s t A : tm X) (B : tm (ext X)) :
    has_type Ctm Cty s (pi A B) ->
    has_type Ctm Cty t A ->
    has_type Ctm Cty (app s t) (beta B t)
| of_lam (A : tm X) (s B : tm (ext X)) :
    @has_type (ext X) (ext Y)
              (var (bound _) .: (Ctm \; map (@shift _)))
              (map (@shift _) A .: (Cty \; (map (@shift _)))) s B ->
    has_type Ctm Cty (lam s) (pi A B).
(*| of_pi (s : tm X) (t : tm (ext X)) :
    has_type*)

Lemma has_type_map {X1 X2 Y1 Y2} (f : X1 -> X2) (g : Y1 -> Y2)
      (Ctm : Y1 -> tm X1) (Cty : Y1 -> tm X1)
      (Dtm : Y2 -> tm X2) (Dty : Y2 -> tm X2)
  :
    (forall i : Y1, map f (Ctm i) = Dtm (g i)) ->
    (forall i : Y1, map f (Cty i) = Dty (g i)) ->
    forall s A,
    has_type Ctm Cty s A ->
    has_type Dtm Dty (map f s) (map f A).
Proof.
  move=> H1 H2 s A tp. elim: tp X2 Y2 f g Dtm Dty H1 H2 => {X1 Y1 s A Ctm Cty}.
  - move=> X1 Y1 Ctm Cty i X2 Y2 f g Dtm Dty H1 H2.
    rewrite H1 H2. exact: of_var.
  - move=> X1 Y1 Ctm Cty s t A B _ ih1 _ ih2 X2 Y2 f g Dtm Dty H1 H2 /=.
    have E: map f (beta B t) = beta (map (upr f) B) (map f t) by admit.
    rewrite E. apply: of_app. exact: ih1. exact: ih2.
  - move=> X1 Y1 Ctm Cty A s B _ ih X2 Y2 f g Dtm Dty H1 H2 /=.
    apply: of_lam. rewrite map_comp. apply: (ih _ _ _ (upr g)).
    + case=>//=x. rewrite map_comp. change (@shift _ \; upr f) with (f \; @shift _).
        by rewrite -map_comp H1.
    + case=>//=[|x]. by rewrite map_comp. rewrite map_comp.
      change (@shift _ \; upr f) with (f \; @shift _).
        by rewrite -map_comp H2.
Admitted.

Lemma ctx_weakening {X Y} (Ctm : Y -> tm X) (Cty : Y -> tm X) (s t A B : tm X) :
  has_type Ctm Cty t B ->
  has_type (s .: Ctm) (A .: Cty) t B.
Proof.
  move=> H. have->: t = map id t by admit. have->: B = map id B by admit.
  apply: (@has_type_map _ _ _ _ _ (@shift _)) H.
Admitted.

Lemma weakening {X Y} (Ctm : Y -> tm X) (Cty : Y -> tm X) (s : tm X) (A : tm X) :
  has_type Ctm Cty s A ->
  has_type (Ctm \; map (@shift _)) (Cty \; map (@shift _)) (map (@shift _) s) (map (@shift _) A).
Proof.
  exact: (@has_type_map _ _ _ _ _ id).
Qed.

Lemma has_type_bind {X1 X2 Y1 Y2} (f : X1 -> tm X2)
      (Ctm : Y1 -> tm X1) (Cty : Y1 -> tm X1)
      (Dtm : Y2 -> tm X2) (Dty : Y2 -> tm X2)
  :
    (forall (i : Y1), has_type Dtm Dty (bind f (Ctm i)) (bind f (Cty i))) ->
    forall s A,
      has_type Ctm Cty s A ->
      has_type Dtm Dty (bind f s) (bind f A).
Proof.
  move=> H s A tp. elim: tp X2 Y2 f Dtm Dty H => {X1 Y1 Ctm Cty s A}.
  - move=> X1 Y1 Ctm Cty i X2 Y2 f Dtm Dty H. exact: H.
  - move=> X1 Y1 Ctm Cty s t A B _ ih1 _ ih2 X2 Y2 f Dtm Dty H /=.
    have E: bind f (beta B t) = beta (bind (up f) B) (bind f t) by admit.
    rewrite E. apply: of_app. exact: ih1. exact: ih2.
  - move=> X1 Y1 Ctm Cty A s B _ ih X2 Y2 f Dtm Dty H /=. apply: of_lam.
    apply: ih. case=>/=[|x].
    + have E:(bind (up f) (map (shift (X:=X1)) A)) = map (shift (X:=X2)) (bind f A) by admit.
      rewrite E. eapply (of_var (var (bound X2) .: (Dtm \; map (shift (X:=X2)))) (map (shift (X:=X2)) (bind f A) .: (Dty \; map (shift (X:=X2)))) (@bound _)).
    + have E1: (bind (up f) (map (shift (X:=X1)) (Ctm x))) = map (@shift _) (bind f (Ctm x)) by admit.
      have E2: (bind (up f) (map (shift (X:=X1)) (Cty x))) = map (@shift _) (bind f (Cty x)) by admit.
      rewrite E1 E2. apply: ctx_weakening. apply: weakening. exact: H.
Admitted.
(*      apply: has_type_map. Focus 3. exact: (H x). Unshelve.
      Focus 3. exact (@shift _). cbn. by []. by [].
Admitted.*)

